# ===================================================================
#
# Copyright (c) 2018, Helder Eijs <helderijs@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ===================================================================

import abc

from Cryptodome.Util.py3compat import iter_range, bord, bchr, ABC

from Cryptodome import Random


class IntegerBase(ABC):

    # Conversions
    @abc.abstractmethod
    def __int__(self):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def __repr__(self):
        pass

    @abc.abstractmethod
    def to_bytes(self, block_size=0, byteorder='big'):
        pass

    @staticmethod
    @abc.abstractmethod
    def from_bytes(byte_string, byteorder='big'):
        pass

    # Relations
    @abc.abstractmethod
    def __eq__(self, term):
        pass

    @abc.abstractmethod
    def __ne__(self, term):
        pass

    @abc.abstractmethod
    def __lt__(self, term):
        pass

    @abc.abstractmethod
    def __le__(self, term):
        pass

    @abc.abstractmethod
    def __gt__(self, term):
        pass

    @abc.abstractmethod
    def __ge__(self, term):
        pass

    @abc.abstractmethod
    def __nonzero__(self):
        pass
    __bool__ = __nonzero__

    @abc.abstractmethod
    def is_negative(self):
        pass

    # Arithmetic operations
    @abc.abstractmethod
    def __add__(self, term):
        pass

    @abc.abstractmethod
    def __sub__(self, term):
        pass

    @abc.abstractmethod
    def __mul__(self, factor):
        pass

    @abc.abstractmethod
    def __floordiv__(self, divisor):
        pass

    @abc.abstractmethod
    def __mod__(self, divisor):
        pass

    @abc.abstractmethod
    def inplace_pow(self, exponent, modulus=None):
        pass

    @abc.abstractmethod
    def __pow__(self, exponent, modulus=None):
        pass

    @abc.abstractmethod
    def __abs__(self):
        pass

    @abc.abstractmethod
    def sqrt(self, modulus=None):
        pass

    @abc.abstractmethod
    def __iadd__(self, term):
        pass

    @abc.abstractmethod
    def __isub__(self, term):
        pass

    @abc.abstractmethod
    def __imul__(self, term):
        pass

    @abc.abstractmethod
    def __imod__(self, term):
        pass

    # Boolean/bit operations
    @abc.abstractmethod
    def __and__(self, term):
        pass

    @abc.abstractmethod
    def __or__(self, term):
        pass

    @abc.abstractmethod
    def __rshift__(self, pos):
        pass

    @abc.abstractmethod
    def __irshift__(self, pos):
        pass

    @abc.abstractmethod
    def __lshift__(self, pos):
        pass

    @abc.abstractmethod
    def __ilshift__(self, pos):
        pass

    @abc.abstractmethod
    def get_bit(self, n):
        pass

    # Extra
    @abc.abstractmethod
    def is_odd(self):
        pass

    @abc.abstractmethod
    def is_even(self):
        pass

    @abc.abstractmethod
    def size_in_bits(self):
        pass

    @abc.abstractmethod
    def size_in_bytes(self):
        pass

    @abc.abstractmethod
    def is_perfect_square(self):
        pass

    @abc.abstractmethod
    def fail_if_divisible_by(self, small_prime):
        pass

    @abc.abstractmethod
    def multiply_accumulate(self, a, b):
        pass

    @abc.abstractmethod
    def set(self, source):
        pass

    @abc.abstractmethod
    def inplace_inverse(self, modulus):
        pass

    @abc.abstractmethod
    def inverse(self, modulus):
        pass

    @abc.abstractmethod
    def gcd(self, term):
        pass

    @abc.abstractmethod
    def lcm(self, term):
        pass

    @staticmethod
    @abc.abstractmethod
    def jacobi_symbol(a, n):
        pass

    @staticmethod
    def _tonelli_shanks(n, p):
        """Tonelli-shanks algorithm for computing the square root
        of n modulo a prime p.

        n must be in the range [0..p-1].
        p must be at least even.

        The return value r is the square root of modulo p. If non-zero,
        another solution will also exist (p-r).

        Note we cannot assume that p is really a prime: if it's not,
        we can either raise an exception or return the correct value.
        """

        # See https://rosettacode.org/wiki/Tonelli-Shanks_algorithm

        if n in (0, 1):
            return n

        if p % 4 == 3:
            root = pow(n, (p + 1) // 4, p)
            if pow(root, 2, p) != n:
                raise ValueError("Cannot compute square root")
            return root

        s = 1
        q = (p - 1) // 2
        while not (q & 1):
            s += 1
            q >>= 1

        z = n.__class__(2)
        while True:
            euler = pow(z, (p - 1) // 2, p)
            if euler == 1:
                z += 1
                continue
            if euler == p - 1:
                break
            # Most probably p is not a prime
            raise ValueError("Cannot compute square root")

        m = s
        c = pow(z, q, p)
        t = pow(n, q, p)
        r = pow(n, (q + 1) // 2, p)

        while t != 1:
            for i in iter_range(0, m):
                if pow(t, 2**i, p) == 1:
                    break
            if i == m:
                raise ValueError("Cannot compute square root of %d mod %d" % (n, p))
            b = pow(c, 2**(m - i - 1), p)
            m = i
            c = b**2 % p
            t = (t * b**2) % p
            r = (r * b) % p

        if pow(r, 2, p) != n:
            raise ValueError("Cannot compute square root")

        return r

    @classmethod
    def random(cls, **kwargs):
        """Generate a random natural integer of a certain size.

        :Keywords:
          exact_bits : positive integer
            The length in bits of the resulting random Integer number.
            The number is guaranteed to fulfil the relation:

                2^bits > result >= 2^(bits - 1)

          max_bits : positive integer
            The maximum length in bits of the resulting random Integer number.
            The number is guaranteed to fulfil the relation:

                2^bits > result >=0

          randfunc : callable
            A function that returns a random byte string. The length of the
            byte string is passed as parameter. Optional.
            If not provided (or ``None``), randomness is read from the system RNG.

        :Return: a Integer object
        """

        exact_bits = kwargs.pop("exact_bits", None)
        max_bits = kwargs.pop("max_bits", None)
        randfunc = kwargs.pop("randfunc", None)

        if randfunc is None:
            randfunc = Random.new().read

        if exact_bits is None and max_bits is None:
            raise ValueError("Either 'exact_bits' or 'max_bits' must be specified")

        if exact_bits is not None and max_bits is not None:
            raise ValueError("'exact_bits' and 'max_bits' are mutually exclusive")

        bits = exact_bits or max_bits
        bytes_needed = ((bits - 1) // 8) + 1
        significant_bits_msb = 8 - (bytes_needed * 8 - bits)
        msb = bord(randfunc(1)[0])
        if exact_bits is not None:
            msb |= 1 << (significant_bits_msb - 1)
        msb &= (1 << significant_bits_msb) - 1

        return cls.from_bytes(bchr(msb) + randfunc(bytes_needed - 1))

    @classmethod
    def random_range(cls, **kwargs):
        """Generate a random integer within a given internal.

        :Keywords:
          min_inclusive : integer
            The lower end of the interval (inclusive).
          max_inclusive : integer
            The higher end of the interval (inclusive).
          max_exclusive : integer
            The higher end of the interval (exclusive).
          randfunc : callable
            A function that returns a random byte string. The length of the
            byte string is passed as parameter. Optional.
            If not provided (or ``None``), randomness is read from the system RNG.
        :Returns:
            An Integer randomly taken in the given interval.
        """

        min_inclusive = kwargs.pop("min_inclusive", None)
        max_inclusive = kwargs.pop("max_inclusive", None)
        max_exclusive = kwargs.pop("max_exclusive", None)
        randfunc = kwargs.pop("randfunc", None)

        if kwargs:
            raise ValueError("Unknown keywords: " + str(kwargs.keys))
        if None not in (max_inclusive, max_exclusive):
            raise ValueError("max_inclusive and max_exclusive cannot be both"
                         " specified")
        if max_exclusive is not None:
            max_inclusive = max_exclusive - 1
        if None in (min_inclusive, max_inclusive):
            raise ValueError("Missing keyword to identify the interval")

        if randfunc is None:
            randfunc = Random.new().read

        norm_maximum = max_inclusive - min_inclusive
        bits_needed = cls(norm_maximum).size_in_bits()

        norm_candidate = -1
        while not 0 <= norm_candidate <= norm_maximum:
            norm_candidate = cls.random(
                                    max_bits=bits_needed,
                                    randfunc=randfunc
                                    )
        return norm_candidate + min_inclusive

    @staticmethod
    @abc.abstractmethod
    def _mult_modulo_bytes(term1, term2, modulus):
        """Multiply two integers, take the modulo, and encode as big endian.
        This specialized method is used for RSA decryption.

        Args:
          term1 : integer
            The first term of the multiplication, non-negative.
          term2 : integer
            The second term of the multiplication, non-negative.
          modulus: integer
            The modulus, a positive odd number.
        :Returns:
            A byte string, with the result of the modular multiplication
            encoded in big endian mode.
            It is as long as the modulus would be, with zero padding
            on the left if needed.
        """
        pass



# -*- coding: utf-8 -*-

# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/api/http.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15google/api/http.proto\x12\ngoogle.api"T\n\x04Http\x12#\n\x05rules\x18\x01 \x03(\x0b\x32\x14.google.api.HttpRule\x12\'\n\x1f\x66ully_decode_reserved_expansion\x18\x02 \x01(\x08"\x81\x02\n\x08HttpRule\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\r\n\x03get\x18\x02 \x01(\tH\x00\x12\r\n\x03put\x18\x03 \x01(\tH\x00\x12\x0e\n\x04post\x18\x04 \x01(\tH\x00\x12\x10\n\x06\x64\x65lete\x18\x05 \x01(\tH\x00\x12\x0f\n\x05patch\x18\x06 \x01(\tH\x00\x12/\n\x06\x63ustom\x18\x08 \x01(\x0b\x32\x1d.google.api.CustomHttpPatternH\x00\x12\x0c\n\x04\x62ody\x18\x07 \x01(\t\x12\x15\n\rresponse_body\x18\x0c \x01(\t\x12\x31\n\x13\x61\x64\x64itional_bindings\x18\x0b \x03(\x0b\x32\x14.google.api.HttpRuleB\t\n\x07pattern"/\n\x11\x43ustomHttpPattern\x12\x0c\n\x04kind\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\tBg\n\x0e\x63om.google.apiB\tHttpProtoP\x01ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\xa2\x02\x04GAPIb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "google.api.http_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\016com.google.apiB\tHttpProtoP\001ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\242\002\004GAPI"
    _globals["_HTTP"]._serialized_start = 37
    _globals["_HTTP"]._serialized_end = 121
    _globals["_HTTPRULE"]._serialized_start = 124
    _globals["_HTTPRULE"]._serialized_end = 381
    _globals["_CUSTOMHTTPPATTERN"]._serialized_start = 383
    _globals["_CUSTOMHTTPPATTERN"]._serialized_end = 430
# @@protoc_insertion_point(module_scope)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from opentelemetry.sdk._logs._internal import (
    LogDroppedAttributesWarning,
    Logger,
    LoggerProvider,
    LoggingHandler,
    LogLimits,
    LogRecordDroppedAttributesWarning,
    LogRecordLimits,
    LogRecordProcessor,
    ReadableLogRecord,
    ReadWriteLogRecord,
)

__all__ = [
    "Logger",
    "LoggerProvider",
    "LoggingHandler",
    "LogLimits",
    "LogRecordLimits",
    "LogRecordProcessor",
    "LogDroppedAttributesWarning",
    "LogRecordDroppedAttributesWarning",
    "ReadableLogRecord",
    "ReadWriteLogRecord",
]



"""
Allow flow-based timers inside Thorium formulas.

These timers keep state across multiple Thorium evaluation cycles, which makes
them useful for cooldowns, delayed reactions, and rate limiting.
"""

import time


def hold(name, seconds):
    """
    Wait for a given period of time, then fire a result of True, requiring
    this state allows for an action to be blocked for evaluation based on
    time.

    A common pattern is to require a ``check.*`` state first and then require
    this timer from the action state so the action only runs after the hold
    period has elapsed.

    USAGE:

    .. code-block:: yaml

        hold_on_a_moment:
          timer.hold:
            - seconds: 30

        cooldown:
          timer.hold:
            - seconds: 900
            - require:
              - check: repeated_failures
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    start = time.time()
    if "timer" not in __context__:
        __context__["timer"] = {}
    if name not in __context__["timer"]:
        __context__["timer"][name] = start
    if (start - __context__["timer"][name]) > seconds:
        ret["result"] = True
        __context__["timer"][name] = start
    return ret



"""
RaftService — lifecycle manager for the Raft node inside a Salt master.

Responsibilities
----------------
* Construct and own the ``Node`` instance.
* Wire ``AsyncTimeoutScheduler`` to the asyncio event loop already running
  inside ``MasterPubServerChannel._publish_daemon``.
* Build one ``SaltPeer`` per entry in ``opts["cluster_peers"]``, keyed to
  the per-peer ``PublishServer`` pushers that ``_publish_daemon`` already
  created.
* Construct ``RaftDispatcher`` and hand it to
  ``MasterPubServerChannel._raft_dispatcher`` so that
  ``handle_pool_publish`` can route inbound Raft RPCs.
* Start the Raft election timer and, when elected leader, drive periodic
  heartbeats.
* Handle dynamic peer joins: when a new master completes the Salt-level
  cluster join, ``notify_peer_joined`` adds it as a non-voting learner.
  The leader will automatically promote it to voter once its log catches up.

Threading / concurrency model
------------------------------
``RaftService`` is created and runs entirely inside the
``EventPublisher`` subprocess owned by ``MasterPubServerChannel``.  All
methods that touch the ``Node`` are called from the asyncio event loop
that Tornado wraps; the Raft core remains synchronous.

Usage (inside ``_publish_daemon``)
-----------------------------------
::

    service = RaftService(opts, aio_loop, peer_pushers)
    service.attach(channel)   # sets channel._raft_dispatcher
    service.start()           # begins election timer
"""

import logging

from salt.cluster.consensus.peer import RaftDispatcher, SaltPeer
from salt.cluster.consensus.raft import AsyncTimeoutScheduler, Node
from salt.cluster.consensus.raft.log import (
    RING_MEMBERS_VOTERS,
    RING_STATUS_ACTIVE,
    RING_STATUS_DESTROYED,
    RING_STATUS_VALID,
    LogEntryType,
    RingConfigStateMachine,
    RingRegistryStateMachine,
    RoutingStateMachine,
)
from salt.cluster.consensus.raft.node import NodeState
from salt.cluster.consensus.storage import SaltStorage

log = logging.getLogger(__name__)

# Heartbeat interval sent by the leader (seconds).  Must be well below the
# follower election timeout floor (~0.15 s per gettimeout defaults).
_HEARTBEAT_INTERVAL = 0.05


class RaftService:
    """
    Owns the Raft ``Node`` for one Salt master process.

    :param opts:         Salt master opts dict.
    :param loop:         The *asyncio* event loop running in this process.
    :param peer_pushers: ``dict[peer_addr, PublishServer]`` - the pushers
                         ``_publish_daemon`` already created, keyed by the
                         peer's interface address (``opts["cluster_peers"]``
                         entry).
    """

    def __init__(self, opts, loop, peer_pushers, voting=True, on_ready=None):
        self.opts = opts
        self.loop = loop
        self._peer_pushers = dict(peer_pushers)  # addr -> PublishServer (mutable copy)
        self._on_ready = on_ready

        # Use the interface address as the Raft node-id so it matches the
        # keys in cluster_peers and the peer_pushers dict.  opts["id"] is
        # the hostname which remote masters do not share; the interface
        # address is the consistent cluster-wide identity.
        node_id = opts["interface"]
        storage = SaltStorage(node_id, opts)
        # voting=False means this node joined dynamically and must wait for a
        # CONFIG entry from the leader before participating in elections.
        #
        # The default Node election window of 150–300 ms is too tight for
        # multi-process Salt masters running over real sockets — a single
        # delayed heartbeat in CI can cause a follower to step up and fight
        # the existing leader for the term.  At ``_HEARTBEAT_INTERVAL`` =
        # 50 ms the rule of thumb is election >= 10× heartbeat, so default
        # to 750–1500 ms here.  ``cluster_election_min`` /
        # ``cluster_election_max`` opts let deployments tune further.
        election_min = opts.get("cluster_election_min", 750)
        election_max = opts.get("cluster_election_max", 1500)
        # ``cluster_max_log_size`` (default ``None``) gates Raft log
        # compaction.  When unset, the log keeps every committed entry
        # forever and snapshots never fire — fine for small clusters
        # but pathological for any long-running deployment.  When set,
        # the membership SM round-trips through the snapshot envelope
        # (raft.snapshot.v1); CONSENSUS_BUGS.md #1's fix and the
        # reconcile_membership hook ensure peer state survives.
        max_log_size = opts.get("cluster_max_log_size")
        # ``cluster_max_voters`` (default ``None``) caps how many peers
        # the leader will auto-promote out of the learner pool.  When
        # the cap is hit, additional joiners stay as non-voting log
        # replicas indefinitely.
        max_voters = opts.get("cluster_max_voters")
        self._node = Node(
            node_id,
            storage=storage,
            voting=voting,
            _follower_min=election_min,
            _follower_max=election_max,
            max_log_size=max_log_size,
            max_voters=max_voters,
        )
        # ``_nodes`` is the multi-ring registry: keys are Raft group
        # ids, values are the local ``Node`` instances.  Slice 1 only
        # carries the main cluster group; later slices spawn per-ring
        # Nodes here as the cluster log commits RING_REGISTRY entries.
        # ``self._node`` (singular) stays as a convenience handle to
        # the cluster Node — every existing call site reads it through
        # that name.
        self._nodes = {"cluster": self._node}
        self._scheduler = AsyncTimeoutScheduler(loop=loop)
        self._node.register_schedule_timeout(self._scheduler.schedule)

        # Wire the membership SM's on_change so we can fire on_ready once
        # this node appears in the committed voter set.
        self._node.membership_sm.on_change = self._on_membership_change

        # Ring registry SM: cluster-log inventory of named rings (one
        # per shardable cache).  Slice 2 of the multi-ring rollout —
        # bringing up / tearing down per-ring Raft groups is wired in
        # slice 3; for now we just track and persist the registry so
        # an operator can record the desired topology.
        self._ring_registry_sm = RingRegistryStateMachine(
            on_change=self._on_ring_registry_change
        )
        self._node.log.register_state_machine(
            "ring_registry_sm", self._ring_registry_sm
        )

        # Routing SM: cluster-log data-type -> ring mapping (e.g.
        # "jobs" -> "jobs_ring", "events" -> None for broadcast).
        # Gate sites consult the routing table once it's populated;
        # absent entries mean broadcast, preserving today's behaviour.
        self._routing_sm = RoutingStateMachine(on_change=self._on_route_change)
        self._node.log.register_state_machine("routing_sm", self._routing_sm)

        # ``Node.__init__`` already loaded any snapshot, but only the
        # state machines registered at construction time
        # (``membership_sm``) saw the restore.  Re-run the restore
        # now that ``ring_sm`` / ``ring_registry_sm`` / ``routing_sm``
        # are registered so a master coming back from a snapshot
        # rebuilds them from disk instead of starting empty (which
        # would be silently wrong under log compaction).
        if storage is not None:
            try:
                snap = storage.load_snapshot()
            except Exception:  # pylint: disable=broad-except
                snap = None
            if snap and "data" in snap:
                self._node.log.restore_state_machines_from_data(snap["data"])

        # The snapshot restore above populated the registry but did
        # not fire its ``on_change`` (``restore_snapshot`` is a pure
        # store).  Drive bring-up for any active rings that this
        # master was hosting before the restart so the per-ring
        # ``Node`` instances reattach to the dispatcher.  Idempotent
        # — ``_bring_up_ring`` no-ops when the ring is already up.
        for ring_id in self._ring_registry_sm.active_rings():
            entry = self._ring_registry_sm.get(ring_id) or {}
            self._on_ring_registry_change(
                ring_id,
                entry.get("founding_voters", []),
                entry.get("status", RING_STATUS_ACTIVE),
            )

        # Build SaltPeer objects - one per cluster peer (all voting at start).
        peers = [
            SaltPeer(addr, pusher, node_id) for addr, pusher in peer_pushers.items()
        ]
        self._node.peers = peers

        # Register a peer factory so Node.on_config_change can create SaltPeers
        # for addresses that appear in CONFIG entries (covers learner->voter path).
        self._node.register_peer_factory(self._make_peer)

        # peer_pushers is keyed by interface address, matching the
        # callback_node field written into RPC envelopes.  Hand the
        # dispatcher the full ``_nodes`` dict so that inbound RPCs are
        # routed by ``raft_group_id``; passing ``self._nodes``
        # (instead of a bare ``self._node``) means subsequent
        # ``register_ring_node`` calls are visible to the dispatcher
        # without extra plumbing.
        self._dispatcher = RaftDispatcher(self._nodes, node_id, self._peer_pushers)

        # If Node started from a saved snapshot the membership SM was
        # restored before this on_change wiring existed, so the cluster-
        # ready / peer-table side effects never fired.  Reconcile now so
        # _on_ready and on_config_change run for the restored view.
        self._node.reconcile_membership()

        self._heartbeat_handle = None
        # Voter health / auto-replacement state (Ongaro thesis §6.4).
        # ``_recently_demoted`` is leader-local; a new leader on failover
        # starts with an empty cooldown table and re-derives unhealthy
        # peers from incoming AppendEntries replies.
        self._recently_demoted = {}
        self._voter_health_handle = None

    # ------------------------------------------------------------------
    # Membership change / readiness
    # ------------------------------------------------------------------

    def _on_membership_change(self, voters, learners):
        """
        Called by ``MembershipStateMachine`` after every committed CONFIG entry.

        Fires ``on_ready`` (once) when this node's address first appears in the
        committed voter set, signalling that it is a full participant and may
        begin serving minion/CLI traffic.

        The default ``"cluster"`` ring is kept in lock-step with the
        cluster voter set so pre-multi-ring callers
        (``ring_membership.owns(opts, key)`` with no ring name) see a
        meaningful answer.  Per-ring sharding is driven separately by
        each per-ring ``Node``'s own ``MembershipStateMachine``.
        """
        # Lazy import keeps the consensus package independent of the
        # ring module's load order.
        import salt.cluster.ring_membership  # pylint: disable=import-outside-toplevel

        # Keep the default "cluster" ring populated with the current
        # voter set.  Empty ``voters`` shouldn't happen in steady
        # state but we tolerate it (an empty ring answers True for
        # every owns() — broadcast semantics).
        if voters:
            salt.cluster.ring_membership.rebuild(list(voters))

        if self._on_ready is None:
            return
        node_id = self._node.node_id
        if node_id in voters:
            log.info(
                "RaftService: node %s is now a committed voter — marking cluster ready",
                node_id,
            )
            self._on_ready()
            self._on_ready = None  # fire only once

    def _on_ring_registry_change(self, ring_id, founding_voters, status):
        """
        Called by :class:`RingRegistryStateMachine` after each
        committed ``RING_REGISTRY`` entry.

        * ``status="active"`` and this master is in ``founding_voters``:
          bring up the per-ring Raft group inside this process.  The
          per-ring ``Node`` shares the same asyncio loop, scheduler,
          and peer transport as the cluster group; only the on-disk
          state and Raft state machines are independent.
        * ``status="destroyed"``: tear down the per-ring Node and
          drop it from the dispatcher's routing table.
        * Otherwise (e.g. an active entry that doesn't list this
          master): nothing to do locally.  Other masters will own the
          ring.
        """
        log.info(
            "RaftService: ring registry committed — ring=%s status=%s "
            "founding_voters=%s",
            ring_id,
            status,
            list(founding_voters or []),
        )
        if status == RING_STATUS_DESTROYED:
            self._tear_down_ring(ring_id)
            return
        if self._node.node_id not in (founding_voters or []):
            # Not a founder — this ring's data plane runs elsewhere.
            return
        self._bring_up_ring(ring_id, founding_voters)

    def _bring_up_ring(self, ring_id, founding_voters):
        """
        Construct and register the per-ring Raft ``Node`` for this
        master.  Idempotent: if the ring is already up locally the
        call is a no-op.

        The per-ring Node has its own :class:`SaltStorage` keyed by
        ``ring_id``, its own ``MembershipStateMachine`` and
        ``RingConfigStateMachine`` registered on its log, its own
        :class:`SaltPeer` instances stamping ``raft_group_id`` into
        outbound RPCs, and its own election + heartbeat path driven
        by the shared scheduler.
        """
        if ring_id == "cluster":
            log.warning(
                "RaftService: refusing to bring up a ring named 'cluster' — "
                "reserved for the main cluster Raft group"
            )
            return
        if ring_id in self._nodes:
            log.debug(
                "RaftService: ring %s already up locally, skipping bring-up",
                ring_id,
            )
            return

        node_id = self._node.node_id
        log.info(
            "RaftService: bringing up ring %s with founders=%s",
            ring_id,
            sorted(founding_voters or []),
        )
        storage = SaltStorage(node_id, self.opts, ring_id=ring_id)
        # Election windows reuse the same opts as the cluster Node —
        # if the operator tuned them for their environment, the
        # per-ring nodes inherit the tuning automatically.
        election_min = self.opts.get("cluster_election_min", 750)
        election_max = self.opts.get("cluster_election_max", 1500)
        ring_node = Node(
            node_id,
            storage=storage,
            voting=True,
            _follower_min=election_min,
            _follower_max=election_max,
            max_log_size=self.opts.get("cluster_max_log_size"),
            max_voters=self.opts.get("cluster_max_voters"),
        )
        ring_node.register_schedule_timeout(self._scheduler.schedule)
        # Per-ring RingConfigStateMachine: each ring has its own
        # members/replicas policy, persisted in its own snapshot
        # envelope.
        ring_config_sm = RingConfigStateMachine(
            on_change=lambda m, r, _ring_id=ring_id: self._on_ring_config_change_for(
                _ring_id, m, r
            )
        )
        ring_node.log.register_state_machine("ring_sm", ring_config_sm)
        # Wire the per-ring MembershipStateMachine's on_change so a
        # committed CONFIG entry on the ring's log triggers a rebuild
        # of the named HashRing.  Without this, ring policy =
        # ``"voters"`` would see the ring's voter set change without
        # the local HashRing reflecting it.
        ring_node.membership_sm.on_change = lambda voters, learners, _ring_id=ring_id: self._on_ring_membership_change_for(
            _ring_id, voters, learners
        )
        # ``Node.__init__`` ran the snapshot restore for
        # ``membership_sm`` only.  Replay it now that ``ring_sm`` is
        # registered so a master that previously hosted this ring
        # picks up its committed policy from disk instead of starting
        # at the SM default.
        try:
            snap = storage.load_snapshot()
        except Exception:  # pylint: disable=broad-except
            snap = None
        if snap and "data" in snap:
            ring_node.log.restore_state_machines_from_data(snap["data"])

        # Build peers for this ring: every founder other than self.
        peers = []
        for addr in sorted(founding_voters or []):
            if addr == node_id:
                continue
            peers.append(self._make_peer(addr, voting=True, raft_group_id=ring_id))
        ring_node.peers = peers
        # Peer factory for membership-change-driven additions.
        ring_node.register_peer_factory(
            lambda addr, voting=True, _ring_id=ring_id: self._make_peer(
                addr, voting=voting, raft_group_id=_ring_id
            )
        )

        # Replay any persisted state.  ``SaltStorage`` was loaded by
        # ``Node.__init__`` already; reconcile fires on_change to
        # rebuild peer flags.
        ring_node.reconcile_membership()

        self._nodes[ring_id] = ring_node
        # Mirror the registration into the dispatcher's routing
        # table so inbound RPCs tagged with this ``raft_group_id``
        # land on the new Node.  The dispatcher keeps its own dict
        # (constructed from a copy at start time) — pushing the
        # update explicitly keeps the two in sync without coupling
        # them.
        self._dispatcher.register_node(ring_id, ring_node)

        ring_node.become_follower()

    def _tear_down_ring(self, ring_id):
        """
        Stop the per-ring ``Node`` and drop it from ``self._nodes``.
        Idempotent.  On-disk state is left in place so an operator
        who re-creates the ring with the same id and founders can
        recover historical state.
        """
        if ring_id == "cluster":
            log.warning("RaftService: refusing to tear down the cluster Raft group")
            return
        ring_node = self._nodes.pop(ring_id, None)
        if ring_node is None:
            return
        self._dispatcher.unregister_node(ring_id)
        # Drop the per-process ring snapshot so subsequent
        # ``owns_for`` calls treat this master as a non-member of the
        # destroyed ring.
        import salt.cluster.ring_membership  # pylint: disable=import-outside-toplevel

        salt.cluster.ring_membership.drop_ring(ring_id)
        log.info("RaftService: tearing down ring %s", ring_id)
        try:
            ring_node.become_follower()
            # Cancel any pending timers stored on the Node; the
            # shared scheduler will simply not re-arm them since the
            # Node is no longer in ``self._nodes`` and the heartbeat
            # tick skips unregistered Nodes.
            if getattr(ring_node, "_follower_timeout", None):
                ring_node._follower_timeout.cancel()
                ring_node._follower_timeout = None
            if getattr(ring_node, "_leader_beacon_timeout", None):
                ring_node._leader_beacon_timeout.cancel()
                ring_node._leader_beacon_timeout = None
        except Exception:  # pylint: disable=broad-except
            log.exception("RaftService: error stopping ring %s", ring_id)

    def _on_ring_membership_change_for(self, ring_id, voters, learners):
        """
        Called by a per-ring :class:`MembershipStateMachine` after each
        committed CONFIG entry on the *ring's* log.  Mirrors the
        cluster-side ``_on_membership_change`` but routes the rebuild
        to the named ring.  Re-applies the current ring policy so the
        ``HashRing`` reflects the new voter set.
        """
        ring_node = self._nodes.get(ring_id)
        if ring_node is None:
            return
        ring_sm = ring_node.log._extra_state_machines.get("ring_sm")
        if ring_sm is None:
            return
        self._on_ring_config_change_for(ring_id, ring_sm.members, ring_sm.replicas)

    def _on_ring_config_change_for(self, ring_id, members, replicas):
        """
        Called by a per-ring :class:`RingConfigStateMachine` after a
        ``RING_CONFIG`` commit on that ring's own log.

        Rebuilds the named ring's :class:`HashRing` from the per-ring
        Raft group's committed voter set.  ``"self"`` means the
        local master is the only ring node (broadcast within the
        ring); ``"voters"`` means the ring contains every committed
        voter and the gate sites will hash by ring ownership.
        """
        import salt.cluster.ring_membership  # pylint: disable=import-outside-toplevel

        ring_node = self._nodes.get(ring_id)
        if ring_node is None:
            log.debug(
                "RaftService: ring=%s policy commit observed but local "
                "Node is not up — skipping rebuild",
                ring_id,
            )
            return
        log.info(
            "RaftService: ring=%s policy committed — members=%s replicas=%d",
            ring_id,
            members,
            replicas,
        )
        if members == RING_MEMBERS_VOTERS:
            voters = ring_node.membership_sm.current_voters()
            salt.cluster.ring_membership.rebuild(ring_id, voters, replicas=replicas)
        else:  # RING_MEMBERS_SELF (default)
            salt.cluster.ring_membership.rebuild(
                ring_id, [ring_node.node_id], replicas=replicas
            )

    def _on_route_change(self, data_type, ring_id):
        """
        Called by :class:`RoutingStateMachine` after each committed
        ``ROUTE`` entry.  Updates the process-local routing snapshot
        consulted by the gate sites in :mod:`salt.master` so a route
        flip takes effect on every master without IPC.
        """
        import salt.cluster.ring_membership  # pylint: disable=import-outside-toplevel

        salt.cluster.ring_membership.set_route(data_type, ring_id)
        if ring_id is None:
            log.info(
                "RaftService: route cleared — data_type=%s now broadcasts",
                data_type,
            )
        else:
            log.info(
                "RaftService: route committed — data_type=%s -> ring=%s",
                data_type,
                ring_id,
            )

    def propose_ring_create(self, ring_id, founding_voters, status=RING_STATUS_ACTIVE):
        """
        Propose a ``RING_REGISTRY`` entry creating (or marking
        destroyed) the named ring.

        Only valid on the leader.  Slice 2 wires the cluster-log
        replication path; the per-ring Raft group does not actually
        come up until slice 3 attaches its lifecycle to
        ``_on_ring_registry_change``.

        :param ring_id:         Operator-chosen name for the ring.
        :param founding_voters: List of master node-ids that will be
                                this ring's initial voter set.  Sorted
                                deterministically before the entry is
                                appended.
        :param status:          ``"active"`` (default) or
                                ``"destroyed"``.

        :raises ValueError:   if ``ring_id`` is empty or ``status`` is
                              unknown.
        :raises RuntimeError: if this node is not currently the leader.
        """
        if not ring_id:
            raise ValueError("propose_ring_create requires a non-empty ring_id")
        if status not in RING_STATUS_VALID:
            raise ValueError(
                f"Unknown ring status {status!r}; "
                f"expected one of {RING_STATUS_VALID}"
            )
        if self._node.state != NodeState.LEADER:
            raise RuntimeError(
                "propose_ring_create must run on the Raft leader; "
                f"this node is in state {self._node.state}"
            )
        founders = sorted(founding_voters or [])
        cmd = {
            "ring_id": ring_id,
            "founding_voters": founders,
            "status": status,
        }
        self._node.log_add(cmd, entry_type=LogEntryType.RING_REGISTRY)

    def propose_ring_destroy(self, ring_id):
        """
        Propose a ``RING_REGISTRY`` entry marking *ring_id* as
        destroyed.  Idempotent at the registry level; the on_change
        callback decides whether to tear down a per-ring Node based
        on its previous status.

        The command omits ``founding_voters`` so the registry SM
        preserves the original founder list as audit history — the
        operator can still see who founded a ring after it's been
        destroyed.
        """
        if not ring_id:
            raise ValueError("propose_ring_destroy requires a non-empty ring_id")
        if self._node.state != NodeState.LEADER:
            raise RuntimeError(
                "propose_ring_destroy must run on the Raft leader; "
                f"this node is in state {self._node.state}"
            )
        cmd = {"ring_id": ring_id, "status": RING_STATUS_DESTROYED}
        self._node.log_add(cmd, entry_type=LogEntryType.RING_REGISTRY)

    def propose_route(self, data_type, ring_id):
        """
        Propose a ``ROUTE`` entry mapping *data_type* to *ring_id*.

        Pass ``ring_id=None`` to clear the route, returning the data
        type to broadcast.  Only valid on the leader.

        :param data_type: Logical cache identifier (e.g. ``"jobs"``).
        :param ring_id:   Ring name to route to, or ``None`` for
                          broadcast.
        :raises ValueError:   if ``data_type`` is empty.
        :raises RuntimeError: if this node is not currently the leader.
        """
        if not data_type:
            raise ValueError("propose_route requires a non-empty data_type")
        if self._node.state != NodeState.LEADER:
            raise RuntimeError(
                "propose_route must run on the Raft leader; "
                f"this node is in state {self._node.state}"
            )
        cmd = {"data_type": data_type, "ring_id": ring_id}
        self._node.log_add(cmd, entry_type=LogEntryType.ROUTE)

    # ------------------------------------------------------------------
    # Peer factory (used by Node.on_config_change)
    # ------------------------------------------------------------------

    def _make_peer(self, addr, voting=True, raft_group_id="cluster"):
        """
        Create a ``SaltPeer`` for *addr*.

        If we already have a pusher for that address (from ``_peer_pushers``)
        it is reused; otherwise we create a new one using the cluster port.
        Called by :meth:`Node.on_config_change` when a CONFIG log entry is
        applied and a previously unknown address appears in the voter list.

        :param raft_group_id: Which Raft group this peer belongs to.
                              Defaults to ``"cluster"`` so the existing
                              cluster-Node call site is unchanged; per-
                              ring bring-up passes the ring's id so the
                              peer stamps every outbound RPC with the
                              ring's group id (and the dispatcher on the
                              receiver routes to the right Node).
        """
        import salt.transport.tcp  # pylint: disable=import-outside-toplevel

        pusher = self._peer_pushers.get(addr)
        if pusher is None:
            port = self.opts.get("cluster_port", 55596)
            pusher = salt.transport.tcp.PublishServer(
                self.opts,
                pull_host=addr,
                pull_port=port,
            )
            self._peer_pushers[addr] = pusher
            # Keep the dispatcher's pusher table in sync.
            self._dispatcher._pushers[addr] = pusher
        return SaltPeer(
            addr,
            pusher,
            self.opts["interface"],
            voting=voting,
            raft_group_id=raft_group_id,
        )

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def attach(self, channel):
        """
        Wire this service into ``MasterPubServerChannel``.

        Sets ``channel._raft_dispatcher`` so that ``handle_pool_publish``
        will route ``cluster/raft/*`` messages here.
        """
        channel._raft_dispatcher = self._dispatcher

    def start(self):
        """
        Start the Raft node as a follower and arm the election timer.

        Must be called from within the running asyncio event loop.
        """
        log.info(
            "RaftService: starting node %s with %d peer(s)",
            self._node.node_id,
            len(self._node.peers),
        )
        self._node.become_follower()
        self._schedule_heartbeat()
        self._schedule_voter_health_check()

    def stop(self):
        """Cancel scheduled callbacks and step the node down."""
        if self._heartbeat_handle is not None:
            self._heartbeat_handle.cancel()
            self._heartbeat_handle = None
        if self._voter_health_handle is not None:
            self._voter_health_handle.cancel()
            self._voter_health_handle = None
        log.info("RaftService: stopped node %s", self._node.node_id)

    # ------------------------------------------------------------------
    # Voter health watchdog (Ongaro thesis §6.4 single-server changes)
    # ------------------------------------------------------------------

    def _schedule_voter_health_check(self):
        """Re-arm the periodic ``_check_voter_health`` timer."""
        interval = self.opts.get("cluster_voter_health_check_interval", 1.0)
        self._voter_health_handle = self._scheduler.schedule(
            interval, self._check_voter_health
        )

    def _check_voter_health(self):
        """
        Periodic leader-side watchdog, run once per scheduled tick
        across every Raft group hosted in this process (cluster +
        per-ring).

        For each group where this master is the current leader: walk
        the voter set, flag voters whose ``last_contact`` is older
        than ``cluster_voter_timeout``, and — if
        ``cluster_auto_replace_voters`` is True — propose a single
        demotion + replacement promotion per tick (Ongaro thesis §6.4
        single-server change semantics).

        Per-group state:

        * ``self._recently_demoted`` is keyed by ``(group_id,
          peer_id)`` so a cooldown on one ring doesn't bleed into
          another.
        * The on-disk sentinel ``cachedir/cluster-health.json`` is a
          structured document with one entry per group — read by
          ``cluster.members`` for the cluster group; per-ring
          consumers can pull from the same file via ``rings.<id>``.

        Safety: relies on each ``Node.propose_voter_demotion`` to
        enforce the ``cluster_min_voters`` floor.  The watchdog
        never bypasses that.  Idempotent on re-entry — a demotion
        CONFIG already in flight leaves the demoted peer in
        ``current_voters`` until commit, so the precondition check
        inside ``propose_voter_demotion`` deduplicates.
        """
        try:
            self._voter_health_handle = None
            now = self._node.get_now()
            timeout = self.opts.get("cluster_voter_timeout", 10.0)
            cooldown = self.opts.get("cluster_demote_cooldown", 60.0)
            auto = self.opts.get("cluster_auto_replace_voters", False)
            min_voters = self.opts.get("cluster_min_voters", 3)

            # Expire cooldown entries first so a fresh tick can
            # promote a candidate that's just exited its cooldown.
            self._recently_demoted = {
                key: ts
                for key, ts in self._recently_demoted.items()
                if now - ts < cooldown
            }

            per_group_unhealthy = {}
            for group_id, node in list(self._nodes.items()):
                if node is None:
                    continue
                unhealthy = []
                if node.state == NodeState.LEADER:
                    voters = set(node.membership_sm.current_voters())
                    for peer in node.peers:
                        if peer.node_id not in voters:
                            continue
                        last = node._peer_last_contact.get(peer.node_id)
                        if last is None:
                            continue
                        if now - last > timeout:
                            unhealthy.append(peer.node_id)
                per_group_unhealthy[group_id] = unhealthy

            self._write_health_sentinel(per_group_unhealthy)

            if not auto:
                return

            for group_id, unhealthy in per_group_unhealthy.items():
                if not unhealthy:
                    continue
                node = self._nodes.get(group_id)
                if node is None or node.state != NodeState.LEADER:
                    continue
                # One demotion + replacement per group per tick.
                target = unhealthy[0]
                if node.propose_voter_demotion(target, min_voters=min_voters):
                    self._recently_demoted[(group_id, target)] = now
                    learners = node.membership_sm.current_learners()
                    for candidate in learners:
                        if (group_id, candidate) in self._recently_demoted:
                            continue
                        if node.match_index.get(candidate, -1) < node.log.index:
                            continue
                        node.propose_voter_promotion_to_replace(candidate)
                        break
        except Exception:  # pylint: disable=broad-except
            log.exception("RaftService: voter health check failed")
        finally:
            self._schedule_voter_health_check()

    def _write_health_sentinel(self, per_group_unhealthy):
        """
        Persist a structured health sentinel covering every Raft
        group hosted in this process.

        Shape (a single JSON document at
        ``cachedir/cluster-health.json``)::

            {
              "updated_at": <epoch>,
              "unhealthy_voters": [<cluster_group_unhealthy>, …],
              "recently_demoted": [<cluster_group_demoted>, …],
              "rings": {
                "<ring_id>": {
                  "unhealthy_voters": […],
                  "recently_demoted": […],
                },
                …
              }
            }

        The top-level ``unhealthy_voters`` / ``recently_demoted``
        fields preserve the pre-multi-ring shape so the
        ``cluster.members`` runner reads them unchanged.  Per-ring
        consumers reach into ``rings.<id>`` for the same view per
        ring.
        """
        import json  # pylint: disable=import-outside-toplevel
        import os  # pylint: disable=import-outside-toplevel
        import time  # pylint: disable=import-outside-toplevel

        import salt.utils.atomicfile  # pylint: disable=import-outside-toplevel

        cachedir = self.opts.get("cachedir")
        if not cachedir:
            return
        path = os.path.join(cachedir, "cluster-health.json")

        # Split the cooldown table by group for the per-ring view.
        cooldown_by_group = {}
        for (group_id, peer_id), _ts in self._recently_demoted.items():
            cooldown_by_group.setdefault(group_id, []).append(peer_id)

        cluster_unhealthy = per_group_unhealthy.get("cluster", [])
        cluster_cooldown = cooldown_by_group.get("cluster", [])

        rings = {}
        for group_id, unhealthy in per_group_unhealthy.items():
            if group_id == "cluster":
                continue
            rings[group_id] = {
                "unhealthy_voters": sorted(unhealthy),
                "recently_demoted": sorted(cooldown_by_group.get(group_id, [])),
            }

        body = {
            "updated_at": time.time(),
            "unhealthy_voters": sorted(cluster_unhealthy),
            "recently_demoted": sorted(cluster_cooldown),
            "rings": rings,
        }
        # Atomic write so the every-N-seconds watchdog rewrite
        # never overlaps an operator's ``cluster.members`` read on
        # a torn-mid-write file.
        try:
            with salt.utils.atomicfile.atomic_open(path, "w") as fp:
                json.dump(body, fp)
        except OSError as exc:
            log.warning(
                "RaftService: could not write health sentinel %s: %s", path, exc
            )

    # ------------------------------------------------------------------
    # Operator overrides
    # ------------------------------------------------------------------

    def propose_voter_demotion(self, peer_id):
        """
        Operator-facing entry point to demote a voter manually.

        Works regardless of ``cluster_auto_replace_voters`` so an operator
        can force a known-bad voter out of the set even when auto-
        replacement is disabled.  Returns the same ``bool`` as
        ``Node.propose_voter_demotion``.

        IPC story: this method runs inside the publish daemon's process,
        which is the only place the ``Node`` is reachable today.  A
        future runner -> daemon command channel can call this directly;
        until then it is callable from python hooks running inside the
        master process.
        """
        log.info("RaftService: operator-requested demotion of %s", peer_id)
        min_voters = self.opts.get("cluster_min_voters", 3)
        return self._node.propose_voter_demotion(peer_id, min_voters=min_voters)

    def propose_voter_promotion(self, peer_id):
        """
        Operator-facing entry point to promote a learner to voter.

        Same operator-override semantics as ``propose_voter_demotion``.
        Subject to the existing ``cluster_max_voters`` cap and the
        caught-up precondition; both are enforced inside
        ``Node.propose_voter_promotion_to_replace``.
        """
        log.info("RaftService: operator-requested promotion of %s", peer_id)
        return self._node.propose_voter_promotion_to_replace(peer_id)

    def notify_peer_joined(self, peer_addr):
        """
        Called when a new master completes the Salt cluster join handshake.

        Adds *peer_addr* as a **non-voting learner** in the Raft cluster.
        If this node is the current leader it immediately starts replicating
        to the learner; once the learner's log catches up the leader will
        automatically propose a CONFIG entry promoting it to voter
        (see :meth:`Node.append_entries_reply`).

        If this node is a follower the learner peer is added to its peer
        list so it will receive ``AppendEntries`` from whichever node
        eventually becomes leader.

        :param peer_addr: The joining master's interface address - the same
                          value used as ``join_peer_id`` in the
                          ``cluster/peer/join-notify`` envelope.
        """
        node_id = self._node.node_id
        if peer_addr == node_id:
            # This is the joining node learning about itself - nothing to do
            # from the peer perspective; our own voting status is controlled
            # by CONFIG entries from the leader.
            return

        existing = {p.node_id for p in self._node.peers}
        if peer_addr in existing:
            log.debug(
                "RaftService: peer %s already known, skipping notify_peer_joined",
                peer_addr,
            )
            return

        log.info("RaftService: adding learner peer %s", peer_addr)

        if self._node.state == NodeState.LEADER:
            # Commit the founding CONFIG *before* the new learner becomes
            # a known peer.  ``_maybe_commit_founding_config`` derives its
            # voter set from ``self._node.peers`` — if the new learner is
            # already in that list it would be incorrectly inducted into
            # the founding voter pool.  No-op when log.index >= 0.
            self._maybe_commit_founding_config()

        learner_peer = self._make_peer(peer_addr, voting=False)
        self._node.peers.append(learner_peer)

        if self._node.state == NodeState.LEADER:
            # Initialise replication tracking for the new learner.
            self._node.next_index[peer_addr] = self._node.log.index + 1
            self._node.match_index[peer_addr] = -1
            # Persist the learner registration in a CONFIG entry so that a
            # subsequent leader failover preserves the learner roster.  Without
            # this, new leaders rebuild ``peers`` only from the committed
            # voter+learner sets in the membership SM; a learner that was only
            # added in the previous leader's in-memory state would disappear
            # and its subsequent RPC replies would trip CandidacyError ("X is
            # not a peer") on the new leader.
            #
            # The cap on ``cluster_max_voters`` applies separately to the
            # *promotion* CONFIG that fires once the learner catches up;
            # this entry only registers the node as a learner, not a voter.
            #
            # Source the voter / learner sets from the leader's in-memory
            # peer list rather than ``membership_sm.current_voters()``:
            # ``on_config_change`` updates peer flags eagerly on the leader
            # when ``log_add`` is called, but ``membership_sm.apply()``
            # only fires after commit.  So immediately after the founding
            # CONFIG is appended (but before quorum acks come in), the
            # leader's ``peers`` already reflect the new view while the SM
            # still has the empty pre-commit state.  Reading from peers
            # avoids the gap.
            from salt.cluster.consensus.raft.log import (
                LogEntryType,  # pylint: disable=import-outside-toplevel
            )

            voters = sorted(
                {self._node.node_id} | {p.node_id for p in self._node.peers if p.voting}
            )
            learners = sorted({p.node_id for p in self._node.peers if not p.voting})
            # Also fold in any committed learners from the SM that may not
            # yet be peer entries (defensive — covers a state-sync restore
            # path).
            for known_learner in self._node.membership_sm.current_learners():
                if known_learner not in voters:
                    learners = sorted(set(learners) | {known_learner})

            # Only emit the CONFIG when this call actually changes the
            # registered set — idempotent against repeat joins.
            current_voters_sm = self._node.membership_sm.current_voters()
            current_learners_sm = self._node.membership_sm.current_learners()
            already_known = (
                peer_addr in current_voters_sm or peer_addr in current_learners_sm
            )
            if not already_known and peer_addr in learners:
                try:
                    self._node.log_add(
                        {"voters": voters, "learners": learners},
                        entry_type=LogEntryType.CONFIG,
                    )
                except Exception:  # pylint: disable=broad-except
                    log.exception(
                        "RaftService: failed to persist learner registration for %s",
                        peer_addr,
                    )
            # Kick off replication immediately.
            self._node.send_append_entries(learner_peer)

    @property
    def node(self):
        return self._node

    @property
    def membership(self):
        """The :class:`~salt.cluster.consensus.raft.log.MembershipStateMachine` for this node."""
        return self._node.membership_sm

    @property
    def dispatcher(self):
        return self._dispatcher

    # ------------------------------------------------------------------
    # Heartbeat
    # ------------------------------------------------------------------

    def _schedule_heartbeat(self):
        self._heartbeat_handle = self.loop.call_later(
            _HEARTBEAT_INTERVAL, self._heartbeat_tick
        )

    def _heartbeat_tick(self):
        """
        Called periodically by the event loop.

        Iterates every Raft group hosted in this process — the
        cluster group plus any per-ring groups — and heartbeats from
        whichever ones consider themselves leader.  On the *first*
        heartbeat after winning an election with an empty log,
        commits a founding CONFIG so the group's voter set is durably
        recorded.

        A single tick services all groups: each group's heartbeat
        load scales with peer count; even a master that hosts a dozen
        rings still issues O(peers) sends per tick.
        """
        try:
            for ring_id, node in list(self._nodes.items()):
                if node is None:
                    continue
                if node.state != NodeState.LEADER:
                    continue
                self._maybe_commit_founding_config(ring_id, node)
                for peer in node.peers:
                    try:
                        ni = node.next_index.get(peer.node_id, node.log.index + 1)
                        # Send a heartbeat (empty) only when the peer
                        # is caught up.  If it's behind, include the
                        # entries so it can advance — important for
                        # lagging learners.
                        entries = [] if ni > node.log.index else None
                        node.send_append_entries(peer, entries=entries)
                    except Exception:  # pylint: disable=broad-except
                        log.exception(
                            "RaftService: error sending heartbeat to %s (ring=%s)",
                            peer.node_id,
                            ring_id,
                        )
        except Exception:  # pylint: disable=broad-except
            log.exception("RaftService: error in heartbeat tick")
        finally:
            self._schedule_heartbeat()

    def _maybe_commit_founding_config(self, ring_id="cluster", node=None):
        """
        Propose the initial CONFIG entry for *node* when its log is
        empty.

        ``ring_id`` defaults to ``"cluster"`` (and ``node`` to
        ``self._node``) so the existing ``self._maybe_commit_founding_config()``
        call sites — heartbeat tick on the cluster Node, the
        ``notify_peer_joined`` flow — keep working without changes.
        Per-ring callers (multi-ring heartbeat tick) pass both.

        Records the founding voter set durably so that a node
        recovering from storage can reconstruct membership without
        relying solely on ``opts["cluster_peers"]`` (cluster group)
        or the registry entry (per-ring groups).

        For the cluster group the bootstrap pool comes from
        ``cluster_peers`` (the static list the operator configured).
        For a per-ring group it comes from that ring's registry
        entry on the cluster log, which records the founding voter
        set chosen by the operator when the ring was created.

        No-ops if the log already has any entries (founding entry
        already written, or this leader inherited a non-empty log).
        """
        if node is None:
            node = self._node
        if node.log.index >= 0:
            return
        # Membership already populated (e.g. via state-sync or test
        # seeding) — no need to synthesize a founding CONFIG.
        if node.membership_sm.current_voters():
            return
        from salt.cluster.consensus.raft.log import (
            LogEntryType,  # pylint: disable=import-outside-toplevel
        )

        if ring_id == "cluster":
            # Deterministic bootstrap pool: sorted set of
            # {this node} ∪ peer addresses.  Every prospective
            # founder runs this code, but only the deterministic
            # founder (lowest interface in the pool; see
            # ``salt/master.py:920`` and
            # ``salt/channel/server.py:2101``) actually writes the
            # CONFIG.
            bootstrap_pool = sorted({node.node_id, *[p.node_id for p in node.peers]})
        else:
            # Per-ring founders come from the registry entry on the
            # cluster log — the operator-chosen founding voters.
            registry_entry = self._ring_registry_sm.get(ring_id) or {}
            bootstrap_pool = sorted(registry_entry.get("founding_voters", []))
            if not bootstrap_pool:
                # Registry entry already missing or destroyed — don't
                # commit a spurious founding CONFIG.
                return

        # ``cluster_max_voters`` (default ``None``) caps the founding
        # voter set.  Excess peers go into the learner set in the
        # same CONFIG entry so they're still durably registered.
        max_voters = self.opts.get("cluster_max_voters")
        if max_voters is not None and len(bootstrap_pool) > max_voters:
            voters = bootstrap_pool[:max_voters]
            learners = bootstrap_pool[max_voters:]
        else:
            voters = bootstrap_pool
            learners = []
        log.info(
            "RaftService: committing founding CONFIG entry (ring=%s) "
            "voters=%s learners=%s",
            ring_id,
            voters,
            learners,
        )
        try:
            node.log_add(
                {"voters": voters, "learners": learners},
                entry_type=LogEntryType.CONFIG,
            )
        except Exception:  # pylint: disable=broad-except
            log.exception("RaftService: failed to propose founding CONFIG entry")


def build_peer_pushers(opts, pushers_list):
    """
    Convert the flat ``pushers`` list from ``_publish_daemon`` into the
    ``dict[addr, PublishServer]`` that ``RaftService`` expects.

    ``_publish_daemon`` builds ``self.pushers`` as a plain list of
    ``PublishServer`` objects in the same order as
    ``opts["cluster_peers"]``.  This helper pairs them back up.

    :param opts:          Salt master opts.
    :param pushers_list:  ``self.pushers`` from ``_publish_daemon``.
    :returns:             ``{peer_addr: PublishServer}``
    """
    peers = opts.get("cluster_peers", [])
    return dict(zip(peers, pushers_list))



"""
OptsDict: Copy-on-write dictionary optimized for Salt's opts pattern.

This module provides a memory-efficient alternative to copy.deepcopy(opts)
by implementing true copy-on-write semantics at the key level, with full
mutation tracking for future unwinding.

Key Features:
- Only duplicates data for keys that are actually mutated
- Tracks all mutations with stack traces for auditing
- Provides reports to identify unwinding opportunities
- Maintains dict interface for backward compatibility
- Thread-safe for concurrent access

Example:
    >>> base_opts = {'grains': {...}, 'pillar': {...}, 'test': False}
    >>> child = OptsDict(base_opts)
    >>> child['test'] = True  # Only 'test' is copied, grains/pillar shared
    >>> child.get_mutation_report()
    {
        'test': {
            'mutated_by': ['salt.utils.state.get_sls_opts:211'],
            'mutation_count': 1,
            'original_value': False,
            'current_value': True
        }
    }
"""

from __future__ import annotations

import copy
import logging
import sys
import threading
import traceback
import weakref
from typing import Any

NamedLoaderContext = None

log = logging.getLogger(__name__)


class _DeletedKey:
    """Sentinel to mark keys that have been deleted from inherited dicts."""

    __slots__ = ()

    def __repr__(self):
        return "<DELETED>"


# Singleton sentinel instance
_DELETED = _DeletedKey()


class DictProxy(dict):
    """
    Proxy for dict that triggers copy-on-write in parent OptsDict on mutation.

    Subclasses dict to pass isinstance checks while providing copy-on-write semantics.
    """

    def __init__(self, target: dict, parent_optsdict: OptsDict, key: str):
        # Initialize underlying dict with target data AND keep _target
        # We need both: underlying dict for C code, _target for our logic
        super().__init__(target)
        object.__setattr__(self, "_target", target)
        object.__setattr__(self, "_parent", parent_optsdict)
        object.__setattr__(self, "_key", key)
        object.__setattr__(self, "_copied", False)

    def _ensure_copied(self):
        """Copy target to parent's _local on first mutation."""
        if not object.__getattribute__(self, "_copied"):
            parent = object.__getattribute__(self, "_parent")
            key = object.__getattribute__(self, "_key")
            target = object.__getattribute__(self, "_target")

            with parent._lock:
                # Deep copy the entire dict to parent's _local
                copied = copy.deepcopy(target)
                parent._local[key] = copied
                object.__setattr__(self, "_target", copied)
                # Also update the underlying dict storage
                dict.clear(self)
                dict.update(self, copied)
                object.__setattr__(self, "_copied", True)

    def __getitem__(self, key):
        target = object.__getattribute__(self, "_target")
        value = target[key]
        # For nested mutable values, copy THIS dict first,
        # then return the value from the copied dict without further wrapping
        if isinstance(value, (dict, list)) and not object.__getattribute__(
            self, "_copied"
        ):
            # Trigger copy-on-access for nested mutables to prevent unwrapped references
            self._ensure_copied()
            # Return value from the now-copied target
            target = object.__getattribute__(self, "_target")
            return target[key]
        return value

    def __setitem__(self, key, value):
        self._ensure_copied()
        object.__getattribute__(self, "_target")[key] = value
        # Also update underlying dict
        dict.__setitem__(self, key, value)

    def __delitem__(self, key):
        self._ensure_copied()
        del object.__getattribute__(self, "_target")[key]
        # Also update underlying dict
        dict.__delitem__(self, key)

    def __iter__(self):
        return iter(object.__getattribute__(self, "_target"))

    def __len__(self):
        return len(object.__getattribute__(self, "_target"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_target"))

    def __str__(self):
        return str(object.__getattribute__(self, "_target"))

    def __contains__(self, key):
        return key in object.__getattribute__(self, "_target")

    def keys(self):
        return object.__getattribute__(self, "_target").keys()

    def values(self):
        return object.__getattribute__(self, "_target").values()

    def items(self):
        return object.__getattribute__(self, "_target").items()

    def get(self, key, default=None):
        return object.__getattribute__(self, "_target").get(key, default)

    def pop(self, key, *args):
        self._ensure_copied()
        result = object.__getattribute__(self, "_target").pop(key, *args)
        # Also update underlying dict
        dict.pop(self, key, *args)
        return result

    def update(self, *args, **kwargs):
        self._ensure_copied()
        object.__getattribute__(self, "_target").update(*args, **kwargs)
        # Also update underlying dict
        dict.update(self, *args, **kwargs)

    def setdefault(self, key, default=None):
        self._ensure_copied()
        result = object.__getattribute__(self, "_target").setdefault(key, default)
        # Also update underlying dict
        dict.setdefault(self, key, default)
        return result

    def __deepcopy__(self, memo):
        """Return a deep copy of the underlying dict, not the proxy."""
        target = object.__getattribute__(self, "_target")
        return copy.deepcopy(target, memo)

    def __reduce_ex__(self, protocol):
        """For pickling, return the underlying dict, not the proxy."""
        target = object.__getattribute__(self, "_target")
        return (dict, (target,))


class ListProxy(list):
    """
    Proxy for list that triggers copy-on-write in parent OptsDict on mutation.

    Subclasses list to pass isinstance checks while providing copy-on-write semantics.
    """

    def __init__(self, target: list, parent_optsdict: OptsDict, key: str):
        # Initialize underlying list with target data AND keep _target
        # We need both: underlying list for C code, _target for our logic
        super().__init__(target)
        object.__setattr__(self, "_target", target)
        object.__setattr__(self, "_parent", parent_optsdict)
        object.__setattr__(self, "_key", key)
        object.__setattr__(self, "_copied", False)

    def _ensure_copied(self):
        """Copy target to parent's _local on first mutation."""
        if not object.__getattribute__(self, "_copied"):
            parent = object.__getattribute__(self, "_parent")
            key = object.__getattribute__(self, "_key")
            target = object.__getattribute__(self, "_target")

            with parent._lock:
                # Deep copy the entire list to parent's _local
                copied = copy.deepcopy(target)
                parent._local[key] = copied
                object.__setattr__(self, "_target", copied)
                object.__setattr__(self, "_copied", True)

    def __getitem__(self, index):
        return object.__getattribute__(self, "_target")[index]

    def __setitem__(self, index, value):
        self._ensure_copied()
        object.__getattribute__(self, "_target")[index] = value

    def __delitem__(self, index):
        self._ensure_copied()
        del object.__getattribute__(self, "_target")[index]

    def __len__(self):
        return len(object.__getattribute__(self, "_target"))

    def __iter__(self):
        return iter(object.__getattribute__(self, "_target"))

    def __contains__(self, item):
        return item in object.__getattribute__(self, "_target")

    def insert(self, index, value):
        self._ensure_copied()
        object.__getattribute__(self, "_target").insert(index, value)

    def append(self, value):
        self._ensure_copied()
        object.__getattribute__(self, "_target").append(value)

    def extend(self, values):
        self._ensure_copied()
        object.__getattribute__(self, "_target").extend(values)

    def remove(self, value):
        self._ensure_copied()
        object.__getattribute__(self, "_target").remove(value)

    def pop(self, index=-1):
        self._ensure_copied()
        return object.__getattribute__(self, "_target").pop(index)

    def __deepcopy__(self, memo):
        """Return a deep copy of the underlying list, not the proxy."""
        return copy.deepcopy(object.__getattribute__(self, "_target"), memo)

    def __reduce_ex__(self, protocol):
        """For pickling, return the underlying list, not the proxy."""
        return (list, (object.__getattribute__(self, "_target"),))

    def clear(self):
        self._ensure_copied()
        object.__getattribute__(self, "_target").clear()

    def __repr__(self):
        return repr(object.__getattribute__(self, "_target"))

    def __str__(self):
        return str(object.__getattribute__(self, "_target"))

    def __eq__(self, other):
        target = object.__getattribute__(self, "_target")
        if isinstance(other, ListProxy):
            return target == object.__getattribute__(other, "_target")
        return target == other

    def index(self, value, *args):
        """Return first index of value."""
        return object.__getattribute__(self, "_target").index(value, *args)

    def count(self, value):
        """Return number of occurrences of value."""
        return object.__getattribute__(self, "_target").count(value)


# DictProxy and ListProxy now subclass dict and list respectively, so they pass
# isinstance checks. This ensures compatibility with Salt code that uses isinstance(x, dict).


class MutationTracker:
    """
    Tracks mutations to OptsDict keys for auditing and future unwinding.

    For each mutated key, we track:
    - Where the mutation happened (file, line, function)
    - When it happened (order of mutations)
    - What the original value was
    - How many times it's been mutated
    """

    def __init__(self, track_mutations: bool = True, max_stack_depth: int = 10):
        self.track_mutations = track_mutations
        self.max_stack_depth = max_stack_depth
        self._lock = threading.RLock()
        if track_mutations:
            self._mutations: dict[str, dict[str, Any]] = {}
            self._mutation_order: list[str] = []
        else:
            # Don't allocate data structures if tracking is disabled
            self._mutations = None
            self._mutation_order = None

    def record_mutation(self, key: str, original_value: Any, new_value: Any):
        """
        Record a mutation of a key with stack trace.

        Args:
            key: The opts key being mutated
            original_value: The original value before mutation
            new_value: The new value being set
        """
        if not self.track_mutations:
            return

        with self._lock:
            # Capture stack trace
            stack = traceback.extract_stack(limit=self.max_stack_depth)
            # Filter out frames from this file
            stack = [frame for frame in stack if "optsdict.py" not in frame.filename]

            # Format stack trace for readability
            caller_info = []
            for frame in stack[-3:]:  # Last 3 frames (most relevant)
                caller_info.append(f"{frame.filename}:{frame.lineno} in {frame.name}")

            if key not in self._mutations:
                # First mutation of this key
                self._mutations[key] = {
                    "original_value": original_value,
                    "current_value": new_value,
                    "mutation_count": 1,
                    "first_mutation_stack": caller_info,
                    "all_mutations": [caller_info],
                    "mutation_sequence": [new_value],
                }
                self._mutation_order.append(key)
            else:
                # Subsequent mutation
                self._mutations[key]["mutation_count"] += 1
                self._mutations[key]["current_value"] = new_value
                self._mutations[key]["all_mutations"].append(caller_info)
                self._mutations[key]["mutation_sequence"].append(new_value)

    def get_mutation_report(self, verbose: bool = False) -> dict[str, Any]:
        """
        Generate a report of all mutations.

        Args:
            verbose: If True, include full mutation history

        Returns:
            Dictionary containing mutation statistics and details
        """
        if not self.track_mutations:
            return {}

        with self._lock:
            if not verbose:
                # Concise report
                return {
                    key: {
                        "mutation_count": info["mutation_count"],
                        "first_mutated_at": info["first_mutation_stack"],
                        "original_value_type": type(info["original_value"]).__name__,
                        "current_value_type": type(info["current_value"]).__name__,
                    }
                    for key, info in self._mutations.items()
                }
            else:
                # Full report with history
                return copy.deepcopy(self._mutations)

    def get_hotspot_keys(self, min_mutations: int = 2) -> list[str]:
        """
        Identify keys that are frequently mutated (hotspots).

        These are candidates for being designed as mutable from the start.

        Args:
            min_mutations: Minimum mutation count to be considered a hotspot

        Returns:
            List of keys sorted by mutation count (descending)
        """
        with self._lock:
            hotspots = [
                (key, info["mutation_count"])
                for key, info in self._mutations.items()
                if info["mutation_count"] >= min_mutations
            ]
            return [
                key for key, _ in sorted(hotspots, key=lambda x: x[1], reverse=True)
            ]

    def get_mutation_locations(self) -> dict[str, set[str]]:
        """
        Get all unique locations where mutations happen.

        Returns:
            Dict mapping location strings to set of keys mutated there
        """
        with self._lock:
            locations = {}
            for key, info in self._mutations.items():
                for stack in info["all_mutations"]:
                    # Use the most recent frame (actual mutation site)
                    location = stack[-1] if stack else "unknown"
                    if location not in locations:
                        locations[location] = set()
                    locations[location].add(key)
            return locations


class OptsDict(dict):
    """
    Copy-on-write dictionary for Salt opts.

    Inherits from dict for full compatibility with isinstance(opts, dict) checks
    while providing copy-on-write semantics for memory efficiency.

    This class implements true copy-on-write semantics at the key level:
    - Keys are only copied when first mutated
    - Nested structures are handled properly (deep COW)
    - Parent data is shared until modification
    - Full mutation tracking for auditing

    Thread-safe for concurrent reads; writes are serialized per instance.

    Example:
        >>> parent_opts = {'grains': {...}, 'pillar': {...}, 'test': False}
        >>> child = OptsDict.from_parent(parent_opts)
        >>> child['test'] = True  # Only copies 'test', not grains/pillar
        >>> 'grains' in child  # True (shared from parent)
        >>> child['test']  # True (local mutation)
        >>> isinstance(child, dict)  # True (for compatibility)
    """

    def __init__(
        self,
        base_dict: dict[str, Any] | None = None,
        parent: OptsDict | None = None,
        track_mutations: bool = False,
        name: str | None = None,
    ):
        """
        Initialize OptsDict.

        Args:
            base_dict: Initial dictionary (for root instance)
            parent: Parent OptsDict to inherit from (for child instances)
            track_mutations: Enable mutation tracking (disabled by default to save memory)
            name: Optional name for debugging (e.g., "loader:states", "state:highstate")
        """
        # Initialize dict parent WITHOUT data (we handle storage ourselves)
        super().__init__()

        self._parent = parent
        self._local = {}  # Keys that have been copied/mutated locally
        self._base = base_dict if base_dict is not None else {}
        self._name = name or f"OptsDict@{id(self)}"
        self._lock = threading.RLock()

        # Mutation tracking
        if parent and parent._tracker:
            # Inherit parent's tracker
            self._tracker = parent._tracker
        else:
            # Root instance - create new tracker
            self._tracker = MutationTracker(track_mutations=track_mutations)

    def _ensure_lock(self) -> threading.RLock:
        """
        Ensure _lock exists and return it.

        This is needed for unpickling, where dict restoration may call
        methods before our attributes are fully initialized.
        """
        if not hasattr(self, "_lock") or self._lock is None:
            object.__setattr__(self, "_lock", threading.RLock())
        return self._lock

    @classmethod
    def from_parent(cls, parent: OptsDict, name: str | None = None) -> OptsDict:
        """
        Create a child OptsDict that shares parent's data.

        Args:
            parent: Parent OptsDict to inherit from
            name: Optional name for debugging

        Returns:
            New OptsDict instance sharing parent's data
        """
        return cls(parent=parent, name=name)

    @classmethod
    def from_dict(
        cls,
        base_dict: dict[str, Any],
        track_mutations: bool = False,
        name: str | None = None,
    ) -> OptsDict:
        """
        Create a root OptsDict from a regular dictionary.

        Args:
            base_dict: Dictionary to wrap
            track_mutations: Enable mutation tracking (disabled by default to save memory)
            name: Optional name for debugging

        Returns:
            New root OptsDict instance
        """
        return cls(base_dict=base_dict, track_mutations=track_mutations, name=name)

    def _get_from_parent_chain(self, key: str) -> tuple[bool, Any]:
        """
        Walk up the parent chain to find a key.

        Returns:
            (found, value) tuple
        """
        current = self._parent
        while current is not None:
            # Check local dict first
            if key in current._local:
                return True, current._local[key]
            # Check base dict if this is a root node
            if current._parent is None and key in current._base:
                return True, current._base[key]
            current = current._parent

        return False, None

    def __getitem__(self, key: str) -> Any:
        """
        Get item with proxy-based copy-on-write for mutable values.

        When accessing mutable values from parent/base, we return a proxy object
        that triggers copy-on-write on first mutation. This provides isolation
        without copying until actually needed.
        """
        with self._ensure_lock():
            # Check local first - if already copied, return direct reference
            if key in self._local:
                value = self._local[key]
                # Check if this key was deleted (masked by sentinel)
                if value is _DELETED:
                    raise KeyError(key)
                return value

            # Check parent chain
            if self._parent is not None:
                found, value = self._get_from_parent_chain(key)
                if found:
                    # Check if this key was deleted in parent chain
                    if value is _DELETED:
                        raise KeyError(key)
                    # Wrap mutable values in proxies to catch mutations
                    if isinstance(value, dict) and not isinstance(value, OptsDict):
                        return DictProxy(value, self, key)
                    elif isinstance(value, list):
                        return ListProxy(value, self, key)
                    # Immutable values can be returned directly
                    return value

            # Check base (root level only)
            if self._parent is None and key in self._base:
                value = self._base[key]
                # Even root instances need proxies to track when values are mutated
                # This allows us to know when a key has been accessed/modified
                if isinstance(value, dict) and not isinstance(value, OptsDict):
                    return DictProxy(value, self, key)
                elif isinstance(value, list):
                    return ListProxy(value, self, key)
                return value

            raise KeyError(key)

    def __setitem__(self, key: str, value: Any):
        """
        Set item with copy-on-write semantics.

        On first write to a key:
        1. Record the mutation
        2. Copy only that key's value to local dict
        3. Store new value
        """
        with self._ensure_lock():
            # Get original value for tracking
            try:
                original_value = self[key]
                is_mutation = True
            except KeyError:
                original_value = None
                is_mutation = False

            # Check if this is the first mutation of this key
            if is_mutation and key not in self._local:
                # First mutation - record it
                self._tracker.record_mutation(key, original_value, value)
            elif key in self._local:
                # Subsequent mutation of already-local key
                self._tracker.record_mutation(key, original_value, value)

            # Store the value locally
            self._local[key] = value

    def _key_in_parent_or_base(self, key: str) -> bool:
        """Check if key exists in parent chain or base dict."""
        # Check parent chain
        if self._parent is not None:
            found, value = self._get_from_parent_chain(key)
            if found and value is not _DELETED:
                return True

        # Check base (root level only)
        if self._parent is None and key in self._base:
            return True

        return False

    def __delitem__(self, key: str):
        """
        Delete item by masking it with a sentinel.

        For keys in local dict: truly delete them.
        For keys in parent/base: mask them with _DELETED sentinel.
        This preserves copy-on-write semantics.
        """
        with self._ensure_lock():
            if key not in self:
                raise KeyError(key)

            if key in self._local:
                # Key is in local - check if it's already deleted
                if self._local[key] is _DELETED:
                    raise KeyError(key)
                # Truly delete if it's ONLY in local (not in parent/base)
                if not self._key_in_parent_or_base(key):
                    del self._local[key]
                else:
                    # Key is in both local and parent/base - mask it
                    self._local[key] = _DELETED
            else:
                # Key is in parent/base only - mask it
                self._local[key] = _DELETED

    def __iter__(self):
        """Iterate over all keys (local + parent chain + base), excluding deleted keys."""
        with self._ensure_lock():
            # Sync underlying dict for C-level iteration (e.g., JSON serialization)
            # This ensures json.dumps() works without needing to_dict()
            # Build items dict first to avoid leaving underlying dict in bad state
            # if an exception occurs during iteration
            items = {}
            for key in self._get_all_keys():
                try:
                    value = self[key]
                    # Skip deleted keys (will raise KeyError)
                    items[key] = value
                except KeyError:
                    # Key was deleted (sentinel), skip it
                    pass

            # Only update underlying dict after successfully building all items
            dict.clear(self)
            for key, value in items.items():
                dict.__setitem__(self, key, value)

            return dict.__iter__(self)

    def _get_all_keys(self):
        """Get all keys from local, parent chain, and base."""
        keys = set(self._local.keys())

        # Add parent chain keys
        if self._parent is not None:
            current = self._parent
            while current is not None:
                keys.update(current._local.keys())
                # Add base keys if this is a root node
                if current._parent is None:
                    keys.update(current._base.keys())
                current = current._parent
        else:
            # This is a root node, add base keys
            keys.update(self._base.keys())

        return keys

    def __len__(self) -> int:
        """Return total number of keys."""
        with self._ensure_lock():
            # Sync underlying dict for C-level access
            _ = iter(self)
            return dict.__len__(self)

    def __contains__(self, key: str) -> bool:
        """Check if key exists in local, parent chain, or base (excluding deleted keys)."""
        with self._ensure_lock():
            # Check if key is in local and not deleted
            if key in self._local:
                return self._local[key] is not _DELETED

            # Check parent chain
            if self._parent is not None:
                found, value = self._get_from_parent_chain(key)
                if found:
                    return value is not _DELETED

            # Check base (root level only)
            if self._parent is None and key in self._base:
                return True

            return False

    def get(self, key: str, default: Any = None) -> Any:
        """Get with default value."""
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key: str, default: Any = None) -> Any:
        """
        Get value for key, setting it to default if not present.

        This is important for OptsDict because dict's setdefault uses C-level
        access which doesn't respect our copy-on-write storage. We need to
        check using __getitem__ and set using __setitem__ to properly trigger
        copy-on-write semantics.
        """
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def pop(self, key: str, *args) -> Any:
        """
        Remove and return value for key.

        This is important for OptsDict because dict's pop uses C-level
        access which doesn't respect our copy-on-write storage.

        For copy-on-write semantics:
        - If key is only in local dict, delete it and return value
        - If key is in parent chain (and possibly also local), unwrap/read the
          value, then mask the key with ``_DELETED`` in ``_local`` so that
          subsequent lookups, iteration and ``in`` checks no longer see it.
          The parent data is never mutated.
        - If key doesn't exist, return default or raise KeyError

        Previously this method short-circuited and returned the value without
        masking, which broke the ``dict.pop`` contract (the key was still
        visible afterwards). Callers that use ``opts.pop("x")`` to both read
        and delete a key would then see it come back during later iteration;
        for example ``salt.cache.mysql_cache._init_client`` relies on pop()
        actually removing ``mysql.*`` entries so that a follow-up
        ``for k in opts: if k.startswith("mysql."):`` loop does not re-emit
        them as connection kwargs (``table_name``, ``fresh_connection``, ...)
        that ``MySQLdb.connect`` then rejects with
        ``TypeError: Connection.__init__() got an unexpected keyword
        argument 'table_name'``.
        """
        with self._ensure_lock():
            if key in self._local:
                value = self._local[key]
                if value is _DELETED:
                    if args:
                        return args[0]
                    raise KeyError(key)
                if self._key_in_parent_or_base(key):
                    # Key is shadowed from parent/base - mask it with the
                    # sentinel so subsequent iteration and ``in`` checks do
                    # not resurrect the parent's value.
                    self._local[key] = _DELETED
                else:
                    del self._local[key]
                return value
            elif self._key_in_parent_or_base(key):
                # Key lives only in parent/base. Read through the normal
                # accessor so mutable values are properly unwrapped, then
                # mark it deleted locally so later iteration/lookups skip
                # it. Parent data is never mutated.
                value = self[key]
                self._local[key] = _DELETED
                return value
            elif args:
                return args[0]
            else:
                raise KeyError(key)

    def keys(self):
        """Return all keys."""
        # Force iteration to sync underlying dict
        _ = iter(self)
        return dict.keys(self)

    def values(self):
        """Return all values."""
        # Force iteration to sync underlying dict
        _ = iter(self)
        return dict.values(self)

    def items(self):
        """Return all items."""
        # Force iteration to sync underlying dict
        _ = iter(self)
        return dict.items(self)

    def copy(self) -> dict[str, Any]:
        """Return a regular dict copy of all data, unwrapping all proxies."""
        with self._ensure_lock():
            result = {}
            for k, v in self.items():
                result[k] = self._unwrap_value(v)
            return result

    def _unwrap_value(self, value):
        """Recursively unwrap DictProxy/ListProxy objects to plain dicts/lists."""
        if isinstance(value, DictProxy):
            # Unwrap DictProxy - get the underlying target and preserve its type
            target = object.__getattribute__(value, "_target")
            # If target is also a DictProxy, we need to get the real underlying dict
            # This shouldn't happen in normal use, but handle it defensively
            while isinstance(target, DictProxy):
                target = object.__getattribute__(target, "_target")
            # Now preserve the type of the real underlying dict (could be OrderedDict, etc.)
            dict_type = (
                type(target) if not isinstance(target, (DictProxy, ListProxy)) else dict
            )
            return dict_type((k, self._unwrap_value(v)) for k, v in value.items())
        elif isinstance(value, ListProxy):
            # Unwrap ListProxy to plain list
            return [self._unwrap_value(item) for item in value]
        elif isinstance(value, dict) and not isinstance(value, OptsDict):
            # Regular dict (including OrderedDict, etc.) - preserve type and recursively unwrap values
            dict_type = type(value)
            return dict_type((k, self._unwrap_value(v)) for k, v in value.items())
        elif isinstance(value, (list, tuple)):
            # Regular list/tuple - recursively unwrap its items
            return type(value)(self._unwrap_value(item) for item in value)
        else:
            # Primitive value or OptsDict - return as-is
            return value

    def to_dict(self) -> dict[str, Any]:
        """Alias for copy() - return regular dict."""
        return self.copy()

    def update(self, other=None, /, **kwargs: Any):  # pylint: disable=arguments-differ
        """Update with another dict.

        Compatible with MutableMapping.update() signature.
        """
        with self._ensure_lock():
            if other:
                for key, value in other.items():
                    self[key] = value
            for key, value in kwargs.items():
                self[key] = value

    # OptsDict-specific methods

    def get_mutation_report(self, verbose: bool = False) -> dict[str, Any]:
        """
        Get report of all mutations across this OptsDict hierarchy.

        Args:
            verbose: Include full mutation history

        Returns:
            Dictionary with mutation statistics
        """
        return self._tracker.get_mutation_report(verbose=verbose)

    def get_local_keys(self) -> set[str]:
        """Return set of keys that have been mutated locally."""
        with self._ensure_lock():
            return set(self._local.keys())

    def get_shared_keys(self) -> set[str]:
        """Return set of keys that are shared from parent/base."""
        with self._ensure_lock():
            all_keys = set(self)
            local_keys = self.get_local_keys()
            return all_keys - local_keys

    def get_root(self) -> OptsDict:
        """
        Get the root OptsDict instance.

        Walks up the parent chain to find the root (the OptsDict with no parent).
        Useful when you need to set a value that should be visible to all children.

        Returns:
            The root OptsDict instance
        """
        root = self
        while root._parent is not None:
            root = root._parent
        return root

    def set_shared(self, key: str, value: Any):
        """
        Set a value on the root OptsDict so all children can see it.

        This is useful for setting shared configuration values (like grains, pillar)
        that need to be visible to all loaders/children, not just the current instance.

        Args:
            key: The configuration key to set
            value: The value to set

        Example:
            opts.set_shared("grains", grains_data)  # All children can now see grains
        """
        root = self.get_root()
        root[key] = value

    def get_memory_stats(self) -> dict[str, Any]:
        """
        Estimate memory usage statistics.

        Returns:
            Dict with local size, shared size, etc.
        """
        with self._ensure_lock():
            import sys

            local_size = sum(
                sys.getsizeof(k) + sys.getsizeof(v) for k, v in self._local.items()
            )

            return {
                "name": self._name,
                "local_keys_count": len(self._local),
                "shared_keys_count": len(self.get_shared_keys()),
                "total_keys_count": len(self),
                "local_size_bytes": local_size,
                "local_size_mb": local_size / (1024 * 1024),
            }

    def log_mutation_summary(self, logger: logging.Logger | None = None):
        """
        Log a summary of mutations for debugging.

        Args:
            logger: Logger to use (defaults to module logger)
        """
        logger = logger or log
        stats = self.get_memory_stats()

        logger.info(
            "OptsDict[%s] memory stats: %d local keys (%.2f MB), %d shared keys",
            self._name,
            stats["local_keys_count"],
            stats["local_size_mb"],
            stats["shared_keys_count"],
        )

        if self._tracker.track_mutations:
            report = self.get_mutation_report(verbose=False)
            if report:
                logger.info("OptsDict[%s] mutations:", self._name)
                for key, info in report.items():
                    logger.info(
                        "  - %s: %d mutations, first at %s",
                        key,
                        info["mutation_count"],
                        (
                            info["first_mutated_at"][-1]
                            if info["first_mutated_at"]
                            else "unknown"
                        ),
                    )

    def __deepcopy__(self, memo):
        """
        Support for copy.deepcopy().

        Uses copy-on-write by creating a child OptsDict instead of deep copying.
        This provides the same isolation semantics as deepcopy while being much
        more memory efficient. When the copied OptsDict is modified, only the
        modified keys are copied, not the entire dict.
        """
        # Use copy-on-write by creating a child OptsDict
        # This shares data with parent until mutations occur
        child = OptsDict.from_parent(self, name=f"{self._name}_deepcopy")
        # Sync underlying dict to support C-level iteration (e.g., JSON serialization)
        # This ensures json.dumps() works without needing to call __iter__() first
        list(child)  # Triggers __iter__() which syncs the underlying dict
        return child

    def __eq__(self, other):
        """
        Compare OptsDict instances based on logical contents.

        dict's built-in __eq__ compares the underlying dict storage directly
        using C-level iteration, which doesn't account for OptsDict's
        copy-on-write storage in _local, _parent, and _base. We need to
        compare the logical key-value pairs instead.
        """
        if not isinstance(other, dict):
            return NotImplemented

        # Get all keys from self
        with self._ensure_lock():
            self_keys = set(self._get_all_keys())

        # For regular dict, just use keys()
        if isinstance(other, OptsDict):
            with other._ensure_lock():
                other_keys = set(other._get_all_keys())
        else:
            other_keys = set(other.keys())

        # Quick check: if key sets differ, not equal
        if self_keys != other_keys:
            return False

        # Compare values for each key
        for key in self_keys:
            if self[key] != other[key]:
                return False

        return True

    def _sync_underlying_dict(self):
        """
        Sync the underlying dict storage with our custom storage.

        This is needed for JSON serialization and other operations that
        bypass our custom __iter__() and use C-level dict iteration.
        """
        with self._ensure_lock():
            # Clear and repopulate the underlying dict
            dict.clear(self)
            for key in self:
                dict.__setitem__(self, key, self[key])

    def __reduce_ex__(self, protocol):
        """
        Custom pickle support for dict subclass.

        When pickling, convert to regular dict and reconstruct via from_dict.
        This avoids issues with dict's unpickling mechanism calling __setitem__
        before __init__ has run.
        """
        # Convert to regular dict
        data = self.to_dict()
        name = getattr(self, "_name", None)
        # Return (callable, args) to reconstruct
        # We pass the name as the third positional arg which maps to name= in from_dict
        return (
            self.__class__.from_dict,
            (data, False, name),  # base_dict, track_mutations=False, name
        )

    def mutate_key(self, key, new_value):
        """
        Update a key in place to preserve cached references.

        This method replaces the pattern of:
            opts[key] = new_dict  # Creates NEW dict, breaks cached references

        With:
            new_dict = ...
            opts.mutate_key(key, new_dict)  # Mutates SAME dict, preserves references

        Why this matters:
        - Loaders cache references to opts["grains"], opts["pillar"], etc.
        - When we replace opts[key] = new_dict, cached refs become stale
        - By mutating in place (clear + update), we preserve object identity
        - Cached references continue to see updates

        Args:
            key: The key to update (e.g., "grains", "pillar")
            new_value: The new value to set

        Example:
            # Instead of:
            opts["grains"] = salt.loader.grains(opts)  # BREAKS cached refs

            # Use:
            new_grains = salt.loader.grains(opts)
            opts.mutate_key("grains", new_grains)  # PRESERVES cached refs
        """
        # Ensure new_value is not None
        if new_value is None:
            # If setting to None, just use regular assignment
            self[key] = new_value
            return

        # Check if key exists and both old and new values are dicts
        if key in self and isinstance(self[key], dict) and isinstance(new_value, dict):
            # Mutate in place to preserve object identity
            self[key].clear()
            self[key].update(new_value)
        else:
            # Initial creation or non-dict value - regular assignment is fine
            self[key] = new_value

    def __repr__(self) -> str:
        """String representation."""
        with self._ensure_lock():
            return (
                f"OptsDict(name={self._name!r}, "
                f"local_keys={len(self._local)}, "
                f"total_keys={len(self)})"
            )


def mutate_opts_key(opts, key, new_value):
    """
    Update an opts key in place to preserve cached references.

    .. deprecated:: 3008.0
        Prefer using the instance method: ``opts.mutate_key(key, value)``
        This function is maintained for backwards compatibility.

    This function replaces the pattern of:
        opts[key] = new_dict  # Creates NEW dict, breaks cached references

    With:
        new_dict = ...
        mutate_opts_key(opts, key, new_dict)  # Mutates SAME dict, preserves references

    Args:
        opts: The opts dict (can be OptsDict or regular dict)
        key: The key to update (e.g., "grains", "pillar")
        new_value: The new value to set

    Example:
        # Preferred (OptsDict instance method):
        opts.mutate_key("grains", new_grains)

        # Still supported (old):
        mutate_opts_key(opts, "grains", new_grains)
    """
    if isinstance(opts, OptsDict):
        # Use instance method
        return opts.mutate_key(key, new_value)

    # Fallback for regular dict (shouldn't happen with new config)
    if new_value is None:
        opts[key] = new_value
        return

    if key in opts and isinstance(opts[key], dict) and isinstance(new_value, dict):
        opts[key].clear()
        opts[key].update(new_value)
    else:
        opts[key] = new_value


def generate_global_mutation_report(include_locations: bool = True) -> str:
    """
    Generate a global report of all mutations across all OptsDict instances.

    This is useful for identifying patterns and unwinding opportunities.

    Args:
        include_locations: Include mutation location analysis

    Returns:
        Formatted report string
    """
    # Note: This would require tracking all OptsDict instances globally
    # For now, this is a placeholder for the concept
    # Could be implemented with a global registry

    report_lines = [
        "=" * 80,
        "OptsDict Global Mutation Report",
        "=" * 80,
        "",
        "This report shows all opts mutations across the Salt minion instance.",
        "Use this to identify unwinding opportunities and refactoring targets.",
        "",
    ]

    # TODO: Implement global tracking
    report_lines.append(
        "(Not implemented yet - call get_mutation_report() on individual instances)"
    )

    return "\n".join(report_lines)


# Global mapping from source dict id() to its OptsDict root
# Keyed by id(source_dict), value is the root OptsDict that wraps it
# Use WeakValueDictionary to allow garbage collection when no longer referenced
_dict_to_root = weakref.WeakValueDictionary()


# Convenience function for backward compatibility
def safe_opts_copy(opts: Any, name: str | None = None) -> OptsDict:
    """
    Create an OptsDict from opts (dict or existing OptsDict).

    This function provides a migration path from copy.deepcopy(opts).

    Args:
        opts: Existing opts (dict or OptsDict)
        name: Optional name for debugging

    Returns:
        OptsDict instance

    Example:
        # OLD:
        opts = copy.deepcopy(opts)

        # NEW:
        from salt.utils.optsdict import safe_opts_copy
        opts = safe_opts_copy(opts, name="loader:states")
    """
    global NamedLoaderContext
    if NamedLoaderContext is None:
        from salt.loader.context import NamedLoaderContext

    if isinstance(opts, NamedLoaderContext):
        opts = opts.value()

    if isinstance(opts, OptsDict):
        # Create child from current opts, not root
        # This ensures the child can see all values in the current opts,
        # including any values set in the current opts' _local dict
        return OptsDict.from_parent(opts, name=name)

    # Converting from regular dict
    # Check if we've already created a root for this specific dict object
    global _dict_to_root  # pylint: disable=global-variable-not-assigned
    opts_id = id(opts)
    if opts_id in _dict_to_root:
        # Reuse existing root - this ensures all OptsDict instances
        # from the same source dict share the same base and see mutations
        root = _dict_to_root[opts_id]
        return OptsDict.from_parent(root, name=name)

    # First time seeing this dict - create a new root that wraps it
    root = OptsDict.from_dict(opts, name=name or f"root@{opts_id}")
    _dict_to_root[opts_id] = root
    return root



"""
Sudo executor module
"""

import shlex

import salt.syspaths
import salt.utils.json
import salt.utils.path

__virtualname__ = "sudo"


def __virtual__():
    if salt.utils.path.which("sudo") and __opts__.get("sudo_user"):
        return __virtualname__
    return False


def execute(opts, data, func, args, kwargs):
    """
    Allow for the calling of execution modules via sudo.

    This module is invoked by the minion if the ``sudo_user`` minion config is
    present.

    Example minion config:

    .. code-block:: yaml

        sudo_user: saltdev

    Once this setting is made, any execution module call done by the minion will be
    run under ``sudo -u <sudo_user> salt-call``.  For example, with the above
    minion config,

    .. code-block:: bash

        salt sudo_minion cmd.run 'cat /etc/sudoers'

    is equivalent to

    .. code-block:: bash

        sudo -u saltdev salt-call cmd.run 'cat /etc/sudoers'

    being run on ``sudo_minion``.
    """
    cmd = [
        "sudo",
        "-u",
        opts.get("sudo_user"),
        "salt-call",
        "--priv",
        opts.get("sudo_user"),
        "--out",
        "json",
        "--metadata",
        "-c",
        opts.get("config_dir"),
        "--",
        data.get("fun"),
    ]
    if data["fun"] in ("state.sls", "state.highstate", "state.apply"):
        kwargs["concurrent"] = True
    for arg in args:
        cmd.append(shlex.quote(str(arg)))
    for key in kwargs:
        cmd.append(shlex.quote(f"{key}={kwargs[key]}"))

    cmd_ret = __salt__["cmd.run_all"](cmd, use_vt=True, python_shell=False)

    if cmd_ret["retcode"] == 0:
        cmd_meta = salt.utils.json.loads(cmd_ret["stdout"])["local"]
        ret = cmd_meta["return"]
        __context__["retcode"] = cmd_meta.get("retcode", 0)
    else:
        ret = cmd_ret["stderr"]
        __context__["retcode"] = cmd_ret["retcode"]

    return ret



"""
Module for sending events using the runner system.

.. versionadded:: 2016.11.0
"""

import logging

import salt.utils.event

log = logging.getLogger(__name__)


def send(tag, data=None):
    """
    Send an event with the given tag and data.

    This is useful for sending events directly to the master from the shell
    with salt-run. It is also quite useful for sending events in orchestration
    states where the ``fire_event`` requisite isn't sufficient because it does
    not support sending custom data with the event.

    Note that event tags will *not* be namespaced like events sent with the
    ``fire_event`` requisite! Whereas events produced from ``fire_event`` are
    prefixed with ``salt/state_result/<jid>/<minion_id>/<name>``, events sent
    using this runner module will have no such prefix. Make sure your reactors
    don't expect a prefix!

    :param tag: the tag to send with the event
    :param data: an optional dictionary of data to send with the event

    CLI Example:

    .. code-block:: bash

        salt-run event.send my/custom/event '{"foo": "bar"}'

    Orchestration Example:

    .. code-block:: yaml

        # orch/command.sls

        run_a_command:
          salt.function:
            - name: cmd.run
            - tgt: my_minion
            - arg:
              - exit {{ pillar['exit_code'] }}

        send_success_event:
          salt.runner:
            - name: event.send
            - tag: my_event/success
            - data:
                foo: bar
            - require:
              - salt: run_a_command

        send_failure_event:
          salt.runner:
            - name: event.send
            - tag: my_event/failure
            - data:
                baz: qux
            - onfail:
              - salt: run_a_command

    .. code-block:: bash

        salt-run state.orchestrate orch.command pillar='{"exit_code": 0}'
        salt-run state.orchestrate orch.command pillar='{"exit_code": 1}'
    """
    data = data or {}
    event = salt.utils.event.get_master_event(
        __opts__, __opts__["sock_dir"], listen=False
    )
    return event.fire_event(data, tag)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Final

K8S_CLUSTER_NAME: Final = "k8s.cluster.name"
"""
The name of the cluster.
"""

K8S_CLUSTER_UID: Final = "k8s.cluster.uid"
"""
A pseudo-ID for the cluster, set to the UID of the `kube-system` namespace.
Note: K8s doesn't have support for obtaining a cluster ID. If this is ever
added, we will recommend collecting the `k8s.cluster.uid` through the
official APIs. In the meantime, we are able to use the `uid` of the
`kube-system` namespace as a proxy for cluster ID. Read on for the
rationale.

Every object created in a K8s cluster is assigned a distinct UID. The
`kube-system` namespace is used by Kubernetes itself and will exist
for the lifetime of the cluster. Using the `uid` of the `kube-system`
namespace is a reasonable proxy for the K8s ClusterID as it will only
change if the cluster is rebuilt. Furthermore, Kubernetes UIDs are
UUIDs as standardized by
[ISO/IEC 9834-8 and ITU-T X.667](https://www.itu.int/ITU-T/studygroups/com17/oid.html).
Which states:

> If generated according to one of the mechanisms defined in Rec.
> ITU-T X.667 | ISO/IEC 9834-8, a UUID is either guaranteed to be
> different from all other UUIDs generated before 3603 A.D., or is
> extremely likely to be different (depending on the mechanism chosen).

Therefore, UIDs between clusters should be extremely unlikely to
conflict.
"""

K8S_CONTAINER_NAME: Final = "k8s.container.name"
"""
The name of the Container from Pod specification, must be unique within a Pod. Container runtime usually uses different globally unique name (`container.name`).
"""

K8S_CONTAINER_RESTART_COUNT: Final = "k8s.container.restart_count"
"""
Number of times the container was restarted. This attribute can be used to identify a particular container (running or stopped) within a container spec.
"""

K8S_CONTAINER_STATUS_LAST_TERMINATED_REASON: Final = (
    "k8s.container.status.last_terminated_reason"
)
"""
Last terminated reason of the Container.
"""

K8S_CONTAINER_STATUS_REASON: Final = "k8s.container.status.reason"
"""
The reason for the container state. Corresponds to the `reason` field of the: [K8s ContainerStateWaiting](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#containerstatewaiting-v1-core) or [K8s ContainerStateTerminated](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#containerstateterminated-v1-core).
"""

K8S_CONTAINER_STATUS_STATE: Final = "k8s.container.status.state"
"""
The state of the container. [K8s ContainerState](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#containerstate-v1-core).
"""

K8S_CRONJOB_ANNOTATION_TEMPLATE: Final = "k8s.cronjob.annotation"
"""
The cronjob annotation placed on the CronJob, the `<key>` being the annotation name, the value being the annotation value.
Note: Examples:

- An annotation `retries` with value `4` SHOULD be recorded as the
  `k8s.cronjob.annotation.retries` attribute with value `"4"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.cronjob.annotation.data` attribute with value `""`.
"""

K8S_CRONJOB_LABEL_TEMPLATE: Final = "k8s.cronjob.label"
"""
The label placed on the CronJob, the `<key>` being the label name, the value being the label value.
Note: Examples:

- A label `type` with value `weekly` SHOULD be recorded as the
  `k8s.cronjob.label.type` attribute with value `"weekly"`.
- A label `automated` with empty string value SHOULD be recorded as
  the `k8s.cronjob.label.automated` attribute with value `""`.
"""

K8S_CRONJOB_NAME: Final = "k8s.cronjob.name"
"""
The name of the CronJob.
"""

K8S_CRONJOB_UID: Final = "k8s.cronjob.uid"
"""
The UID of the CronJob.
"""

K8S_DAEMONSET_ANNOTATION_TEMPLATE: Final = "k8s.daemonset.annotation"
"""
The annotation placed on the DaemonSet, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- A label `replicas` with value `1` SHOULD be recorded
  as the `k8s.daemonset.annotation.replicas` attribute with value `"1"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.daemonset.annotation.data` attribute with value `""`.
"""

K8S_DAEMONSET_LABEL_TEMPLATE: Final = "k8s.daemonset.label"
"""
The label placed on the DaemonSet, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `guestbook` SHOULD be recorded
  as the `k8s.daemonset.label.app` attribute with value `"guestbook"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.daemonset.label.injected` attribute with value `""`.
"""

K8S_DAEMONSET_NAME: Final = "k8s.daemonset.name"
"""
The name of the DaemonSet.
"""

K8S_DAEMONSET_UID: Final = "k8s.daemonset.uid"
"""
The UID of the DaemonSet.
"""

K8S_DEPLOYMENT_ANNOTATION_TEMPLATE: Final = "k8s.deployment.annotation"
"""
The annotation placed on the Deployment, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- A label `replicas` with value `1` SHOULD be recorded
  as the `k8s.deployment.annotation.replicas` attribute with value `"1"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.deployment.annotation.data` attribute with value `""`.
"""

K8S_DEPLOYMENT_LABEL_TEMPLATE: Final = "k8s.deployment.label"
"""
The label placed on the Deployment, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `replicas` with value `0` SHOULD be recorded
  as the `k8s.deployment.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.deployment.label.injected` attribute with value `""`.
"""

K8S_DEPLOYMENT_NAME: Final = "k8s.deployment.name"
"""
The name of the Deployment.
"""

K8S_DEPLOYMENT_UID: Final = "k8s.deployment.uid"
"""
The UID of the Deployment.
"""

K8S_HPA_METRIC_TYPE: Final = "k8s.hpa.metric.type"
"""
The type of metric source for the horizontal pod autoscaler.
Note: This attribute reflects the `type` field of spec.metrics[] in the HPA.
"""

K8S_HPA_NAME: Final = "k8s.hpa.name"
"""
The name of the horizontal pod autoscaler.
"""

K8S_HPA_SCALETARGETREF_API_VERSION: Final = (
    "k8s.hpa.scaletargetref.api_version"
)
"""
The API version of the target resource to scale for the HorizontalPodAutoscaler.
Note: This maps to the `apiVersion` field in the `scaleTargetRef` of the HPA spec.
"""

K8S_HPA_SCALETARGETREF_KIND: Final = "k8s.hpa.scaletargetref.kind"
"""
The kind of the target resource to scale for the HorizontalPodAutoscaler.
Note: This maps to the `kind` field in the `scaleTargetRef` of the HPA spec.
"""

K8S_HPA_SCALETARGETREF_NAME: Final = "k8s.hpa.scaletargetref.name"
"""
The name of the target resource to scale for the HorizontalPodAutoscaler.
Note: This maps to the `name` field in the `scaleTargetRef` of the HPA spec.
"""

K8S_HPA_UID: Final = "k8s.hpa.uid"
"""
The UID of the horizontal pod autoscaler.
"""

K8S_HUGEPAGE_SIZE: Final = "k8s.hugepage.size"
"""
The size (identifier) of the K8s huge page.
"""

K8S_JOB_ANNOTATION_TEMPLATE: Final = "k8s.job.annotation"
"""
The annotation placed on the Job, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- A label `number` with value `1` SHOULD be recorded
  as the `k8s.job.annotation.number` attribute with value `"1"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.job.annotation.data` attribute with value `""`.
"""

K8S_JOB_LABEL_TEMPLATE: Final = "k8s.job.label"
"""
The label placed on the Job, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `jobtype` with value `ci` SHOULD be recorded
  as the `k8s.job.label.jobtype` attribute with value `"ci"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.job.label.automated` attribute with value `""`.
"""

K8S_JOB_NAME: Final = "k8s.job.name"
"""
The name of the Job.
"""

K8S_JOB_UID: Final = "k8s.job.uid"
"""
The UID of the Job.
"""

K8S_NAMESPACE_ANNOTATION_TEMPLATE: Final = "k8s.namespace.annotation"
"""
The annotation placed on the Namespace, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- A label `ttl` with value `0` SHOULD be recorded
  as the `k8s.namespace.annotation.ttl` attribute with value `"0"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.namespace.annotation.data` attribute with value `""`.
"""

K8S_NAMESPACE_LABEL_TEMPLATE: Final = "k8s.namespace.label"
"""
The label placed on the Namespace, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `kubernetes.io/metadata.name` with value `default` SHOULD be recorded
  as the `k8s.namespace.label.kubernetes.io/metadata.name` attribute with value `"default"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.namespace.label.data` attribute with value `""`.
"""

K8S_NAMESPACE_NAME: Final = "k8s.namespace.name"
"""
The name of the namespace that the pod is running in.
"""

K8S_NAMESPACE_PHASE: Final = "k8s.namespace.phase"
"""
The phase of the K8s namespace.
Note: This attribute aligns with the `phase` field of the
[K8s NamespaceStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#namespacestatus-v1-core).
"""

K8S_NODE_ANNOTATION_TEMPLATE: Final = "k8s.node.annotation"
"""
The annotation placed on the Node, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `node.alpha.kubernetes.io/ttl` with value `0` SHOULD be recorded as
  the `k8s.node.annotation.node.alpha.kubernetes.io/ttl` attribute with value `"0"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.node.annotation.data` attribute with value `""`.
"""

K8S_NODE_CONDITION_STATUS: Final = "k8s.node.condition.status"
"""
The status of the condition, one of True, False, Unknown.
Note: This attribute aligns with the `status` field of the
[NodeCondition](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#nodecondition-v1-core).
"""

K8S_NODE_CONDITION_TYPE: Final = "k8s.node.condition.type"
"""
The condition type of a K8s Node.
Note: K8s Node conditions as described
by [K8s documentation](https://v1-32.docs.kubernetes.io/docs/reference/node/node-status/#condition).

This attribute aligns with the `type` field of the
[NodeCondition](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#nodecondition-v1-core)

The set of possible values is not limited to those listed here. Managed Kubernetes environments,
or custom controllers MAY introduce additional node condition types.
When this occurs, the exact value as reported by the Kubernetes API SHOULD be used.
"""

K8S_NODE_LABEL_TEMPLATE: Final = "k8s.node.label"
"""
The label placed on the Node, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `kubernetes.io/arch` with value `arm64` SHOULD be recorded
  as the `k8s.node.label.kubernetes.io/arch` attribute with value `"arm64"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.node.label.data` attribute with value `""`.
"""

K8S_NODE_NAME: Final = "k8s.node.name"
"""
The name of the Node.
"""

K8S_NODE_SYSTEM_CONTAINER_NAME: Final = "k8s.node.system_container.name"
"""
The name of the system container running on the K8s Node.
"""

K8S_NODE_UID: Final = "k8s.node.uid"
"""
The UID of the Node.
"""

K8S_PERSISTENTVOLUME_ANNOTATION_TEMPLATE: Final = (
    "k8s.persistentvolume.annotation"
)
"""
The annotation placed on the PersistentVolume, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `pv.kubernetes.io/provisioned-by` with value `kubernetes.io/aws-ebs` SHOULD be recorded as
  the `k8s.persistentvolume.annotation.pv.kubernetes.io/provisioned-by` attribute with value `"kubernetes.io/aws-ebs"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.persistentvolume.annotation.data` attribute with value `""`.
"""

K8S_PERSISTENTVOLUME_LABEL_TEMPLATE: Final = "k8s.persistentvolume.label"
"""
The label placed on the PersistentVolume, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `type` with value `ssd` SHOULD be recorded as
  the `k8s.persistentvolume.label.type` attribute with value `"ssd"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.persistentvolume.label.data` attribute with value `""`.
"""

K8S_PERSISTENTVOLUME_NAME: Final = "k8s.persistentvolume.name"
"""
The name of the PersistentVolume.
"""

K8S_PERSISTENTVOLUME_RECLAIM_POLICY: Final = (
    "k8s.persistentvolume.reclaim_policy"
)
"""
The reclaim policy of the PersistentVolume.
Note: This attribute aligns with the `persistentVolumeReclaimPolicy` field of the
[K8s PersistentVolumeSpec](https://kubernetes.io/docs/reference/kubernetes-api/config-and-storage-resources/persistent-volume-v1/#PersistentVolumeSpec).
"""

K8S_PERSISTENTVOLUME_STATUS_PHASE: Final = "k8s.persistentvolume.status.phase"
"""
The phase of the PersistentVolume.
Note: This attribute aligns with the `phase` field of the
[K8s PersistentVolumeStatus](https://kubernetes.io/docs/reference/kubernetes-api/config-and-storage-resources/persistent-volume-v1/#PersistentVolumeStatus).
"""

K8S_PERSISTENTVOLUME_UID: Final = "k8s.persistentvolume.uid"
"""
The UID of the PersistentVolume.
"""

K8S_PERSISTENTVOLUMECLAIM_ANNOTATION_TEMPLATE: Final = (
    "k8s.persistentvolumeclaim.annotation"
)
"""
The annotation placed on the PersistentVolumeClaim, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `volume.beta.kubernetes.io/storage-provisioner` with value `kubernetes.io/aws-ebs` SHOULD be recorded as
  the `k8s.persistentvolumeclaim.annotation.volume.beta.kubernetes.io/storage-provisioner` attribute with value `"kubernetes.io/aws-ebs"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.persistentvolumeclaim.annotation.data` attribute with value `""`.
"""

K8S_PERSISTENTVOLUMECLAIM_LABEL_TEMPLATE: Final = (
    "k8s.persistentvolumeclaim.label"
)
"""
The label placed on the PersistentVolumeClaim, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `my-app` SHOULD be recorded as
  the `k8s.persistentvolumeclaim.label.app` attribute with value `"my-app"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.persistentvolumeclaim.label.data` attribute with value `""`.
"""

K8S_PERSISTENTVOLUMECLAIM_NAME: Final = "k8s.persistentvolumeclaim.name"
"""
The name of the PersistentVolumeClaim.
"""

K8S_PERSISTENTVOLUMECLAIM_STATUS_PHASE: Final = (
    "k8s.persistentvolumeclaim.status.phase"
)
"""
The phase of the PersistentVolumeClaim.
Note: This attribute aligns with the `phase` field of the
[K8s PersistentVolumeClaimStatus](https://kubernetes.io/docs/reference/kubernetes-api/config-and-storage-resources/persistent-volume-claim-v1/#PersistentVolumeClaimStatus).
"""

K8S_PERSISTENTVOLUMECLAIM_UID: Final = "k8s.persistentvolumeclaim.uid"
"""
The UID of the PersistentVolumeClaim.
"""

K8S_POD_ANNOTATION_TEMPLATE: Final = "k8s.pod.annotation"
"""
The annotation placed on the Pod, the `<key>` being the annotation name, the value being the annotation value.
Note: Examples:

- An annotation `kubernetes.io/enforce-mountable-secrets` with value `true` SHOULD be recorded as
  the `k8s.pod.annotation.kubernetes.io/enforce-mountable-secrets` attribute with value `"true"`.
- An annotation `mycompany.io/arch` with value `x64` SHOULD be recorded as
  the `k8s.pod.annotation.mycompany.io/arch` attribute with value `"x64"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.pod.annotation.data` attribute with value `""`.
"""

K8S_POD_HOSTNAME: Final = "k8s.pod.hostname"
"""
Specifies the hostname of the Pod.
Note: The K8s Pod spec has an optional hostname field, which can be used to specify a hostname.
Refer to [K8s docs](https://kubernetes.io/docs/concepts/services-networking/dns-pod-service/#pod-hostname-and-subdomain-field)
for more information about this field.

This attribute aligns with the `hostname` field of the
[K8s PodSpec](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.34/#podspec-v1-core).
"""

K8S_POD_IP: Final = "k8s.pod.ip"
"""
IP address allocated to the Pod.
Note: This attribute aligns with the `podIP` field of the
[K8s PodStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.34/#podstatus-v1-core).
"""

K8S_POD_LABEL_TEMPLATE: Final = "k8s.pod.label"
"""
The label placed on the Pod, the `<key>` being the label name, the value being the label value.
Note: Examples:

- A label `app` with value `my-app` SHOULD be recorded as
  the `k8s.pod.label.app` attribute with value `"my-app"`.
- A label `mycompany.io/arch` with value `x64` SHOULD be recorded as
  the `k8s.pod.label.mycompany.io/arch` attribute with value `"x64"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.pod.label.data` attribute with value `""`.
"""

K8S_POD_LABELS_TEMPLATE: Final = "k8s.pod.labels"
"""
Deprecated: Replaced by `k8s.pod.label`.
"""

K8S_POD_NAME: Final = "k8s.pod.name"
"""
The name of the Pod.
"""

K8S_POD_START_TIME: Final = "k8s.pod.start_time"
"""
The start timestamp of the Pod.
Note: Date and time at which the object was acknowledged by the Kubelet.
This is before the Kubelet pulled the container image(s) for the pod.

This attribute aligns with the `startTime` field of the
[K8s PodStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.34/#podstatus-v1-core),
in ISO 8601 (RFC 3339 compatible) format.
"""

K8S_POD_STATUS_PHASE: Final = "k8s.pod.status.phase"
"""
The phase for the pod. Corresponds to the `phase` field of the: [K8s PodStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.33/#podstatus-v1-core).
"""

K8S_POD_STATUS_REASON: Final = "k8s.pod.status.reason"
"""
The reason for the pod state. Corresponds to the `reason` field of the: [K8s PodStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.33/#podstatus-v1-core).
"""

K8S_POD_UID: Final = "k8s.pod.uid"
"""
The UID of the Pod.
"""

K8S_REPLICASET_ANNOTATION_TEMPLATE: Final = "k8s.replicaset.annotation"
"""
The annotation placed on the ReplicaSet, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- A label `replicas` with value `0` SHOULD be recorded
  as the `k8s.replicaset.annotation.replicas` attribute with value `"0"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.replicaset.annotation.data` attribute with value `""`.
"""

K8S_REPLICASET_LABEL_TEMPLATE: Final = "k8s.replicaset.label"
"""
The label placed on the ReplicaSet, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `guestbook` SHOULD be recorded
  as the `k8s.replicaset.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.replicaset.label.injected` attribute with value `""`.
"""

K8S_REPLICASET_NAME: Final = "k8s.replicaset.name"
"""
The name of the ReplicaSet.
"""

K8S_REPLICASET_UID: Final = "k8s.replicaset.uid"
"""
The UID of the ReplicaSet.
"""

K8S_REPLICATIONCONTROLLER_NAME: Final = "k8s.replicationcontroller.name"
"""
The name of the replication controller.
"""

K8S_REPLICATIONCONTROLLER_UID: Final = "k8s.replicationcontroller.uid"
"""
The UID of the replication controller.
"""

K8S_RESOURCEQUOTA_NAME: Final = "k8s.resourcequota.name"
"""
The name of the resource quota.
"""

K8S_RESOURCEQUOTA_RESOURCE_NAME: Final = "k8s.resourcequota.resource_name"
"""
The name of the K8s resource a resource quota defines.
Note: The value for this attribute can be either the full `count/<resource>[.<group>]` string (e.g., count/deployments.apps, count/pods), or, for certain core Kubernetes resources, just the resource name (e.g., pods, services, configmaps). Both forms are supported by Kubernetes for object count quotas. See [Kubernetes Resource Quotas documentation](https://kubernetes.io/docs/concepts/policy/resource-quotas/#quota-on-object-count) for more details.
"""

K8S_RESOURCEQUOTA_UID: Final = "k8s.resourcequota.uid"
"""
The UID of the resource quota.
"""

K8S_SERVICE_ANNOTATION_TEMPLATE: Final = "k8s.service.annotation"
"""
The annotation placed on the Service, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `prometheus.io/scrape` with value `true` SHOULD be recorded as
  the `k8s.service.annotation.prometheus.io/scrape` attribute with value `"true"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.service.annotation.data` attribute with value `""`.
"""

K8S_SERVICE_ENDPOINT_ADDRESS_TYPE: Final = "k8s.service.endpoint.address_type"
"""
The address type of the service endpoint.
Note: The network address family or type of the endpoint.
This attribute aligns with the `addressType` field of the
[K8s EndpointSlice](https://kubernetes.io/docs/reference/kubernetes-api/service-resources/endpoint-slice-v1/).
It is used to differentiate metrics when a Service is backed by multiple address types
(e.g., in dual-stack clusters).
"""

K8S_SERVICE_ENDPOINT_CONDITION: Final = "k8s.service.endpoint.condition"
"""
The condition of the service endpoint.
Note: The current operational condition of the service endpoint.
An endpoint can have multiple conditions set at once (e.g., both `serving` and `terminating` during rollout).
This attribute aligns with the condition fields in the [K8s EndpointSlice](https://kubernetes.io/docs/reference/kubernetes-api/service-resources/endpoint-slice-v1/).
"""

K8S_SERVICE_ENDPOINT_ZONE: Final = "k8s.service.endpoint.zone"
"""
The zone of the service endpoint.
Note: The zone where the endpoint is located, typically corresponding to a failure domain.
This attribute aligns with the `zone` field of endpoints in the
[K8s EndpointSlice](https://kubernetes.io/docs/reference/kubernetes-api/service-resources/endpoint-slice-v1/).
It enables zone-aware monitoring of service endpoint distribution and supports
features like [Topology Aware Routing](https://kubernetes.io/docs/concepts/services-networking/topology-aware-routing/).

If the zone is not populated (e.g., nodes without the `topology.kubernetes.io/zone` label),
the attribute value will be an empty string.
"""

K8S_SERVICE_LABEL_TEMPLATE: Final = "k8s.service.label"
"""
The label placed on the Service, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `my-service` SHOULD be recorded as
  the `k8s.service.label.app` attribute with value `"my-service"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.service.label.data` attribute with value `""`.
"""

K8S_SERVICE_NAME: Final = "k8s.service.name"
"""
The name of the Service.
"""

K8S_SERVICE_PUBLISH_NOT_READY_ADDRESSES: Final = (
    "k8s.service.publish_not_ready_addresses"
)
"""
Whether the Service publishes not-ready endpoints.
Note: Whether the Service is configured to publish endpoints before the pods are ready.
This attribute is typically used to indicate that a Service (such as a headless
Service for a StatefulSet) allows peer discovery before pods pass their readiness probes.
It aligns with the `publishNotReadyAddresses` field of the
[K8s ServiceSpec](https://kubernetes.io/docs/reference/kubernetes-api/service-resources/service-v1/#ServiceSpec).
"""

K8S_SERVICE_SELECTOR_TEMPLATE: Final = "k8s.service.selector"
"""
The selector key-value pair placed on the Service, the `<key>` being the selector key, the value being the selector value.
Note: These selectors are used to correlate with pod labels. Each selector key-value pair becomes a separate attribute.

Examples:

- A selector `app=my-app` SHOULD be recorded as
  the `k8s.service.selector.app` attribute with value `"my-app"`.
- A selector `version=v1` SHOULD be recorded as
  the `k8s.service.selector.version` attribute with value `"v1"`.
"""

K8S_SERVICE_TRAFFIC_DISTRIBUTION: Final = "k8s.service.traffic_distribution"
"""
The traffic distribution policy for the Service.
Note: Specifies how traffic is distributed to endpoints for this Service.
This attribute aligns with the `trafficDistribution` field of the
[K8s ServiceSpec](https://kubernetes.io/docs/reference/networking/virtual-ips/#traffic-distribution).
Known values include `PreferSameZone` (prefer endpoints in the same zone as the client) and
`PreferSameNode` (prefer endpoints on the same node, fallback to same zone, then cluster-wide).
If this field is not set on the Service, the attribute SHOULD NOT be emitted.
When not set, Kubernetes distributes traffic evenly across all endpoints cluster-wide.
"""

K8S_SERVICE_TYPE: Final = "k8s.service.type"
"""
The type of the Kubernetes Service.
Note: This attribute aligns with the `type` field of the
[K8s ServiceSpec](https://kubernetes.io/docs/reference/kubernetes-api/service-resources/service-v1/#ServiceSpec).
"""

K8S_SERVICE_UID: Final = "k8s.service.uid"
"""
The UID of the Service.
"""

K8S_STATEFULSET_ANNOTATION_TEMPLATE: Final = "k8s.statefulset.annotation"
"""
The annotation placed on the StatefulSet, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- A label `replicas` with value `1` SHOULD be recorded
  as the `k8s.statefulset.annotation.replicas` attribute with value `"1"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.statefulset.annotation.data` attribute with value `""`.
"""

K8S_STATEFULSET_LABEL_TEMPLATE: Final = "k8s.statefulset.label"
"""
The label placed on the StatefulSet, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `replicas` with value `0` SHOULD be recorded
  as the `k8s.statefulset.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.statefulset.label.injected` attribute with value `""`.
"""

K8S_STATEFULSET_NAME: Final = "k8s.statefulset.name"
"""
The name of the StatefulSet.
"""

K8S_STATEFULSET_UID: Final = "k8s.statefulset.uid"
"""
The UID of the StatefulSet.
"""

K8S_STORAGECLASS_NAME: Final = "k8s.storageclass.name"
"""
The name of K8s [StorageClass](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/#storageclass-v1-storage-k8s-io) object.
"""

K8S_VOLUME_NAME: Final = "k8s.volume.name"
"""
The name of the K8s volume.
"""

K8S_VOLUME_TYPE: Final = "k8s.volume.type"
"""
The type of the K8s volume.
"""


class K8sContainerStatusReasonValues(Enum):
    CONTAINER_CREATING = "ContainerCreating"
    """The container is being created."""
    CRASH_LOOP_BACK_OFF = "CrashLoopBackOff"
    """The container is in a crash loop back off state."""
    CREATE_CONTAINER_CONFIG_ERROR = "CreateContainerConfigError"
    """There was an error creating the container configuration."""
    ERR_IMAGE_PULL = "ErrImagePull"
    """There was an error pulling the container image."""
    IMAGE_PULL_BACK_OFF = "ImagePullBackOff"
    """The container image pull is in back off state."""
    OOM_KILLED = "OOMKilled"
    """The container was killed due to out of memory."""
    COMPLETED = "Completed"
    """The container has completed execution."""
    ERROR = "Error"
    """There was an error with the container."""
    CONTAINER_CANNOT_RUN = "ContainerCannotRun"
    """The container cannot run."""


class K8sContainerStatusStateValues(Enum):
    TERMINATED = "terminated"
    """The container has terminated."""
    RUNNING = "running"
    """The container is running."""
    WAITING = "waiting"
    """The container is waiting."""


class K8sNamespacePhaseValues(Enum):
    ACTIVE = "active"
    """Active namespace phase as described by [K8s API](https://pkg.go.dev/k8s.io/api@v0.31.3/core/v1#NamespacePhase)."""
    TERMINATING = "terminating"
    """Terminating namespace phase as described by [K8s API](https://pkg.go.dev/k8s.io/api@v0.31.3/core/v1#NamespacePhase)."""


class K8sNodeConditionStatusValues(Enum):
    CONDITION_TRUE = "true"
    """condition_true."""
    CONDITION_FALSE = "false"
    """condition_false."""
    CONDITION_UNKNOWN = "unknown"
    """condition_unknown."""


class K8sNodeConditionTypeValues(Enum):
    READY = "Ready"
    """The node is healthy and ready to accept pods."""
    DISK_PRESSURE = "DiskPressure"
    """Pressure exists on the disk size—that is, if the disk capacity is low."""
    MEMORY_PRESSURE = "MemoryPressure"
    """Pressure exists on the node memory—that is, if the node memory is low."""
    PID_PRESSURE = "PIDPressure"
    """Pressure exists on the processes—that is, if there are too many processes on the node."""
    NETWORK_UNAVAILABLE = "NetworkUnavailable"
    """The network for the node is not correctly configured."""


class K8sPersistentvolumeReclaimPolicyValues(Enum):
    DELETE = "Delete"
    """The volume will be deleted when released from its claim."""
    RECYCLE = "Recycle"
    """The volume will be recycled (basic scrub) when released from its claim."""
    RETAIN = "Retain"
    """The volume will be retained when released from its claim."""


class K8sPersistentvolumeStatusPhaseValues(Enum):
    AVAILABLE = "Available"
    """The volume is available and not yet bound to a claim."""
    BOUND = "Bound"
    """The volume is bound to a claim."""
    FAILED = "Failed"
    """The volume has failed its automatic reclamation."""
    PENDING = "Pending"
    """The volume is being provisioned."""
    RELEASED = "Released"
    """The claim has been deleted but the volume is not yet available."""


class K8sPersistentvolumeclaimStatusPhaseValues(Enum):
    BOUND = "Bound"
    """The claim is bound to a volume."""
    LOST = "Lost"
    """The claim has lost its underlying volume (the volume does not exist anymore)."""
    PENDING = "Pending"
    """The claim has not yet been bound to a volume."""


class K8sPodStatusPhaseValues(Enum):
    PENDING = "Pending"
    """The pod has been accepted by the system, but one or more of the containers has not been started. This includes time before being bound to a node, as well as time spent pulling images onto the host."""
    RUNNING = "Running"
    """The pod has been bound to a node and all of the containers have been started. At least one container is still running or is in the process of being restarted."""
    SUCCEEDED = "Succeeded"
    """All containers in the pod have voluntarily terminated with a container exit code of 0, and the system is not going to restart any of these containers."""
    FAILED = "Failed"
    """All containers in the pod have terminated, and at least one container has terminated in a failure (exited with a non-zero exit code or was stopped by the system)."""
    UNKNOWN = "Unknown"
    """For some reason the state of the pod could not be obtained, typically due to an error in communicating with the host of the pod."""


class K8sPodStatusReasonValues(Enum):
    EVICTED = "Evicted"
    """The pod is evicted."""
    NODE_AFFINITY = "NodeAffinity"
    """The pod is in a status because of its node affinity."""
    NODE_LOST = "NodeLost"
    """The reason on a pod when its state cannot be confirmed as kubelet is unresponsive on the node it is (was) running."""
    SHUTDOWN = "Shutdown"
    """The node is shutdown."""
    UNEXPECTED_ADMISSION_ERROR = "UnexpectedAdmissionError"
    """The pod was rejected admission to the node because of an error during admission that could not be categorized."""


class K8sServiceEndpointAddressTypeValues(Enum):
    IPV4 = "IPv4"
    """IPv4 address type."""
    IPV6 = "IPv6"
    """IPv6 address type."""
    FQDN = "FQDN"
    """FQDN address type."""


class K8sServiceEndpointConditionValues(Enum):
    READY = "ready"
    """The endpoint is ready to receive new connections."""
    SERVING = "serving"
    """The endpoint is currently handling traffic."""
    TERMINATING = "terminating"
    """The endpoint is in the process of shutting down."""


class K8sServiceTypeValues(Enum):
    CLUSTER_IP = "ClusterIP"
    """ClusterIP service type."""
    NODE_PORT = "NodePort"
    """NodePort service type."""
    LOAD_BALANCER = "LoadBalancer"
    """LoadBalancer service type."""
    EXTERNAL_NAME = "ExternalName"
    """ExternalName service type."""


class K8sVolumeTypeValues(Enum):
    PERSISTENT_VOLUME_CLAIM = "persistentVolumeClaim"
    """A [persistentVolumeClaim](https://v1-30.docs.kubernetes.io/docs/concepts/storage/volumes/#persistentvolumeclaim) volume."""
    CONFIG_MAP = "configMap"
    """A [configMap](https://v1-30.docs.kubernetes.io/docs/concepts/storage/volumes/#configmap) volume."""
    DOWNWARD_API = "downwardAPI"
    """A [downwardAPI](https://v1-30.docs.kubernetes.io/docs/concepts/storage/volumes/#downwardapi) volume."""
    EMPTY_DIR = "emptyDir"
    """An [emptyDir](https://v1-30.docs.kubernetes.io/docs/concepts/storage/volumes/#emptydir) volume."""
    SECRET = "secret"
    """A [secret](https://v1-30.docs.kubernetes.io/docs/concepts/storage/volumes/#secret) volume."""
    LOCAL = "local"
    """A [local](https://v1-30.docs.kubernetes.io/docs/concepts/storage/volumes/#local) volume."""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

ELASTICSEARCH_NODE_NAME: Final = "elasticsearch.node.name"
"""
Represents the human-readable identifier of the node/instance to which a request was routed.
"""
